# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import base64
import random
import types
from collections import defaultdict
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING

import networkx as nx
import structlog

from pydiverse.pipedag import ExternalTableReference
from pydiverse.pipedag.context import (
    ConfigContext,
    DAGContext,
    FinalTaskState,
    RunContextServer,
)
from pydiverse.pipedag.context.context import CacheValidationMode
from pydiverse.pipedag.context.trace_hook import TraceHook
from pydiverse.pipedag.core import GroupNode, Result, Stage, Task
from pydiverse.pipedag.core.config import PipedagConfig
from pydiverse.pipedag.core.group_node import BarrierTask, VisualizationStyle
from pydiverse.pipedag.core.stage import CommitStageTask
from pydiverse.pipedag.core.task import TaskGetItem
from pydiverse.pipedag.errors import DuplicateNameError, FlowError

try:
    # pydot should be an optional dependency
    import pydot
except ImportError:
    pydot = types.ModuleType("pydot")
    pydot.Dot = None

if TYPE_CHECKING:
    from pydiverse.pipedag.engine import OrchestrationEngine


class Flow:
    """
    A flow represents a collection of dependent Tasks.

    A flow is defined by using it as a context manager.
    Any stages and tasks defined inside the flow context will automatically get added
    to the flow with the correct dependency wiring.

    :param name: The name of the flow.

    Examples
    --------
    ::

        @materialize
        def my_materializing_task():
            return Table(...)

        with Flow("my_flow") as flow:
            with Stage("stage_1") as stage_1:
                task_1 = my_materializing_task()
                task_2 = another_materializing_task(task_1)

            with Stage("stage_2") as stage_2:
                task_3 = yet_another_task(task_3)
                ...

        result = flow.run()
    """

    def __init__(
        self,
        name: str = "default",
    ):
        self.name = name

        self.logger = structlog.get_logger(logger_name=type(self).__name__)
        self.stages: dict[str, Stage] = {}
        self.tasks: list[Task] = []
        self.group_nodes: list[GroupNode] = []

        self.graph = nx.DiGraph()
        self.explicit_graph: nx.DiGraph | None = None

    def __enter__(self):
        # Check that flows don't get nested
        try:
            DAGContext.get()
        except LookupError:
            pass
        else:
            raise RuntimeError("DAG Context already exists. Flows can't be nested.")

        # Initialize context (both Flow and Stage use DAGContext to transport
        # information to @materialize annotations within the flow and to
        # support nesting of stages)
        self._ctx = DAGContext(flow=self, stage=None, group_node=None)
        self._ctx.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._ctx.__exit__()
        del self._ctx

        self.explicit_graph = self.build_graph()

    def __getitem__(self, name) -> Stage:
        """Retrieves a stage by name.

        :param name: The name of the stage.
        :raises KeyError: if no stage with the specified name exists.
        """

        if stage := self.stages.get(name):
            return stage
        raise KeyError(f"Couldn't find a stage with name '{name}' in flow.")

    def add_stage(self, stage: Stage):
        if stage.name in self.stages:
            raise DuplicateNameError(f"Stage with name '{stage.name}' already exists.")

        stage.id = len(self.stages)
        self.stages[stage.name] = stage

    def add_group_node(self, group_node: GroupNode):
        group_node.id = len(self.group_nodes)
        self.group_nodes.append(group_node)

    def add_task(self, task: Task):
        assert self.stages[task.stage.name] is task.stage

        task.id = len(self.tasks)
        self.tasks.append(task)
        self.graph.add_node(task)

    def add_edge(self, from_: Task, to: Task):
        if from_ not in self.graph:
            raise FlowError(f"Can't add edge from {from_} to {to} because `from` is not in the flow.")
        if to not in self.graph:
            raise FlowError(f"Can't add edge from {from_} to {to} because `to` is not in the flow.")

        self.graph.add_edge(from_, to)

    def build_graph(self) -> nx.DiGraph:
        if not nx.is_directed_acyclic_graph(self.graph):
            raise FlowError("Graph is not a DAG")

        explicit_graph = self.graph.copy()
        stages = self.stages.values()
        group_nodes = self.group_nodes

        # Barrier Tasks
        for group_node in group_nodes:
            if group_node.entry_barrier_task is not None:
                explicit_graph.add_node(group_node.entry_barrier_task)
                # link entry barrier to leaves before
                before_leaves = group_node.prev_tasks
                for task in group_node.prev_tasks:
                    before_leaves = before_leaves - set(task.input_tasks.values())
                for task in before_leaves:
                    explicit_graph.add_edge(task, group_node.entry_barrier_task)
                # link barrier to source tasks within group node
                for task in group_node.tasks:
                    input_tasks = set(task.input_tasks.values())
                    if not (input_tasks & group_node.tasks):
                        explicit_graph.add_edge(group_node.entry_barrier_task, task)
            if group_node.exit_barrier_task is not None:
                all_tasks = set(group_node.outer_stage.tasks)
                after_tasks = all_tasks - group_node.tasks - group_node.prev_tasks
                if after_tasks:
                    explicit_graph.add_node(group_node.exit_barrier_task)
                    # link exit barrier to leaf tasks within group node
                    in_leaves = group_node.tasks
                    for task in group_node.tasks:
                        in_leaves = in_leaves - set(task.input_tasks.values())
                    for task in in_leaves:
                        explicit_graph.add_edge(task, group_node.exit_barrier_task)
                    # link exit barrier to source tasks after group bode
                    for task in after_tasks:
                        input_tasks = set(task.input_tasks.values())
                        if not (input_tasks & after_tasks):
                            explicit_graph.add_edge(group_node.exit_barrier_task, task)

        # Commit Tasks
        commit_tasks: dict[Stage, CommitStageTask] = {}
        for stage in stages:
            commit_tasks[stage] = stage.commit_task
            explicit_graph.add_node(stage.commit_task)

        # Add dependencies
        # Because stages can be nested, we add the commit task dependencies
        # to the leaf nodes of the subgraph
        contained_tasks: defaultdict[Stage, list[Task]] = defaultdict(lambda: [])
        leaf_nodes: defaultdict[Stage, list[Task]] = defaultdict(lambda: [])

        for stage in stages:
            s = stage
            while s is not None:
                contained_tasks[s].extend(stage.tasks)
                s = s.outer_stage

        for stage in stages:
            subgraph = self.graph.subgraph(contained_tasks[stage])  # type: nx.DiGraph
            for task, degree in subgraph.out_degree:  # type: ignore
                if degree != 0:
                    continue
                leaf_nodes[stage].append(task)

        # Add commit task downstream dependencies to leaf nodes
        for stage in stages:
            for leaf in leaf_nodes[stage]:
                commit_task = commit_tasks[stage]
                explicit_graph.add_edge(leaf, commit_task)

        # Add commit task upstream dependencies
        for task in self.tasks:
            for parent, _ in self.graph.in_edges(task):  # type: Task
                if parent.stage == task.stage:
                    continue
                if task.stage.is_inner(parent.stage):
                    continue

                commit_task = commit_tasks[parent.stage]
                explicit_graph.add_edge(commit_task, task)

        # Ensure inner stages get committed before outer stages
        for stage in stages:
            if stage.outer_stage is not None:
                explicit_graph.add_edge(commit_tasks[stage], commit_tasks[stage.outer_stage])

        return explicit_graph

    def get_subflow(self, *components: Task | TaskGetItem | Stage) -> "Subflow":
        from pydiverse.pipedag import Stage, Task
        from pydiverse.pipedag.core.task import TaskGetItem

        tasks = []
        stages = []
        for component in components:
            if isinstance(component, Task):
                tasks.append(component)
            elif isinstance(component, TaskGetItem):
                tasks.append(component.task)
            elif isinstance(component, Stage):
                stages.append(component)
            else:
                raise TypeError(f"Unexpected object of type {type(component).__name__}")

        return Subflow(self, tasks=tasks, stages=stages)

    def run(
        self,
        *components: Task | TaskGetItem | Stage,
        config: ConfigContext = None,
        orchestration_engine: "OrchestrationEngine" = None,
        trace_hook: TraceHook = None,
        fail_fast: bool | None = None,
        cache_validation_mode: CacheValidationMode | None = None,
        disable_cache_function: bool | None = None,
        ignore_task_version: bool | None = None,
        ignore_position_hashes: bool = False,
        inputs: dict[Task, ExternalTableReference] | None = None,
        **kwargs,
    ) -> Result:
        """Execute the flow.
        This will execute the flow and all tasks within using the orchestration engine.

        :param components: An optional selection of tasks or stages that should
            get executed. If no values are provided, all tasks and stages get executed.

            If you specify a subset of tasks, those tasks get executed even when
            they are cache valid, and they don't get committed.

            If you specify a subset of stages, all tasks and substages of those stages
            get executed and committed.
        :param config:
            A :py:class:`ConfigContext` to use as the configuration for executing
            this flow. If no config is provided, then pipedag uses the
            current innermost open ConfigContext, or if no such config exists,
            it gets the default config from :py:attr:`PipedagConfig.default`.
        :param orchestration_engine:
            The orchestration engine to use. If no engine is provided, the
            orchestration engine from the config gets used.
        :param fail_fast:
            Whether exceptions should get raised or swallowed.
            If set to True, exceptions that occur get immediately raised and the
            flow gets aborted.
        :param cache_validation_mode:
            Override the cache validation mode. See :py:class:`CacheValidationMode`.
            For None, the cache validation mode from the config gets used.
            See :doc:`/reference/config` :ref:`section-cache_validation` for more
            information.
        :param disable_cache_function:
            Override the disable_cache_function setting from the config.
            See :doc:`/reference/config` :ref:`section-cache_validation` for more
            information.
        :param ignore_task_version:
            Override the ignore_task_version setting from the config.
            See :doc:`/reference/config` :ref:`section-cache_validation` for more
            information.
        :param ignore_position_hashes:
            If ``True``, the position hashes of tasks are not checked
            when retrieving the inputs of a task from the cache.
            This simplifies execution of subgraphs if you don't care whether inputs to
            that subgraph are cache invalid. This allows multiple modifications in the
            Graph before the next run updating the cache.
            Attention: This may break automatic cache invalidation.
            And for this to work, any task producing an input
            for the chosen subgraph may never be used more
            than once per stage.
            NOTE: This is only supported for the SequentialEngine and SQLTablestore
        :param inputs:
            Optionally provide the outputs for a subset of tasks.
            The format is expected as
            dict[Task, ExternalTableReference].
            When the output of said tasks are retrieved as inputs for another
            dependent task they will be fetched from the ExternalTableReference instead.

            NOTE: This feature is experimental and is not compatible
            with all aspects of pipedag.
            Only ExternalTableReferences are supported as inputs.
            This feature does not work well in combination with Tasks returning
            multiple outputs (nout > 1 or RawSql tasks returning multiple tables).
            In the case of multiple outputs, the provided ExternalTableReference
            will override all outputs of the task.
            It is currently only supported when using
            the SQLTablestore at the moment. Using it disables caching.
        :param kwargs:
            Other keyword arguments that get passed on directly to the
            ``run()`` method of the orchestration engine. Consequently, these
            keyword arguments are engine dependant.
        :return:
            A :py:class:`Result` object for the current flow run.

        Examples
        --------
        .. code-block:: python

            with Flow() as flow:
                with Stage("stage_1") as stage_1:
                    task_1 = ...
                    task_2 = ...

                with Stage("stage_2") as stage_2:
                    task_3 = ...
                    task_4 = ...

            # Execute the entire flow
            flow.run()

            # Execute (and commit) only stage 1
            flow.run(stage_1)

            # Execute (but DON'T commit) only task 1 and task 4
            flow.run(task_1, task_4)

        """
        if ignore_position_hashes:
            self.logger.warn(
                "Using ignore_position_hashes=True! "
                "This may break automatic cache invalidation. "
                "And for this to work, any task producing an input "
                "for the chosen subgraph may never be used more than once per stage."
            )
        subflow = self.get_subflow(*components)

        # Get the ConfigContext to use
        if config is None:
            try:
                config = ConfigContext.get()
            except LookupError:
                config = PipedagConfig.default.get(flow=self.name)

        # update cache_validation_mode
        if cache_validation_mode is None and subflow.is_tasks_subflow:
            # If subflow consists of a subset of tasks (-> not a subset of stages)
            # then we want to skip cache validity checking to ensure the tasks
            # always get executed.
            cache_validation_mode = CacheValidationMode.FORCE_CACHE_INVALID

        # Evolve config using the arguments passed to flow.run
        config = config.evolve(
            fail_fast=(fail_fast if fail_fast is not None else config.fail_fast),
            cache_validation={
                k: v
                for k, v in [
                    ("mode", cache_validation_mode),
                    ("disable_cache_function", disable_cache_function),
                    ("ignore_task_version", ignore_task_version),
                ]
                if v is not None
            },
        )

        if (
            config.cache_validation.mode == CacheValidationMode.NORMAL
            and config.cache_validation.disable_cache_function
        ):
            raise ValueError(
                "disable_cache_function=True is not allowed in combination with "
                f"cache_validation_mode=NORMAL: {config.cache_validation}"
            )

        if trace_hook is None:
            trace_hook = TraceHook()

        with config, RunContextServer(subflow, trace_hook):
            if orchestration_engine is None:
                orchestration_engine = config.create_orchestration_engine()
            result = orchestration_engine.run(subflow, ignore_position_hashes, inputs, **kwargs)

            try:
                visualization_url = result.visualize_url()
                self.logger.info("Flow visualization", url=visualization_url)
            except RuntimeError as e:
                self.logger.info("Flow visualization disabled", reason=str(e))

        trace_hook.run_complete(result)

        if not result.successful and config.fail_fast:
            raise result.exception or Exception("Flow run failed")

        return result

    def get_stage(self, name: str) -> Stage:
        """Retrieves a stage by name.
        Alias for :py:meth:`Flow.__getitem__`.

        :param name: The name of the stage.
        :return: The stage.
        :raises KeyError: if no stage with the specified name exists.
        """

        return self[name]

    # Visualization
    def visualize(self, result: Result | None = None, visualization_tag: str | None = None):
        """Visualizes the flow as a graph.

        If you are running in a jupyter notebook, the graph will get displayed inline.
        Otherwise, it will get rendered to a pdf that then gets opened in your browser.

        Requires `Graphviz <https://graphviz.org>`_ to be installed on your computer.

        :param result: An optional :py:class:`Result` instance.
            If provided, the visualization will contain additional information such
            as which tasks ran successfully, or failed.
        """
        return self.get_subflow().visualize(result, visualization_tag)

    def visualize_url(self, result: Result | None = None, visualization_tag: str | None = None) -> str:
        """Visualizes the flow as a graph and returns a URL to view the visualization.

        If you don't have Graphviz installed on your computer (and thus aren't able to
        use the :py:meth:`~.visualize()` method) then this is the easiest way to
        visualize the flow.
        For this we use a free service called `Kroki <https://kroki.io>`_.

        :param result: An optional :py:class:`Result` instance.
            If provided, the visualization will contain additional information such
            as which tasks ran successfully, or failed.
        :return:
            A URL that, when opened, displays the graph.
        """
        return self.get_subflow().visualize_url(result, visualization_tag)

    def visualize_pydot(self, result: Result | None = None, visualization_tag: str | None = None) -> pydot.Dot:
        """Visualizes the flow as a graph and return a ``pydot.Dot`` graph.

        :param result: An optional :py:class:`Result` instance.
            If provided, the visualization will contain additional information such
            as which tasks ran successfully, or failed.
        :return: A ``pydot.Dot`` graph.
        """
        return self.get_subflow().visualize_pydot(result, visualization_tag)


class Subflow:
    def __init__(self, flow: Flow, tasks: list[Task], stages: list[Stage]):
        self.flow = flow
        self.name = flow.name
        self.is_tasks_subflow = len(tasks) > 0

        if tasks and stages:
            raise ValueError("You can only specify either a subset of tasks OR subset of stages to run, but not both.")
        elif not tasks and not stages:
            self.selected_stages = set(flow.stages.values())
            self.selected_tasks = set(flow.tasks)
            return

        # Preprocessing
        self.selected_stages = set(stages)
        for stage in self.flow.stages.values():
            if stage.outer_stage in self.selected_stages:
                self.selected_stages.add(stage)

        # Construct set of selected tasks
        self.selected_tasks = set(tasks)

        if not tasks:
            for stage in self.selected_stages:
                for task in stage.tasks:
                    self.selected_tasks.add(task)
                self.selected_tasks.add(stage.commit_task)

    def get_tasks(self) -> Iterable[Task]:
        """
        All tasks that should get executed as part of this sub flow.
        Guaranteed to be in the same order as they were added to the flow
        (they will be ordered topologically).
        """
        graph = self.flow.explicit_graph
        for task in self.flow.tasks:
            if task in self.selected_tasks or (
                isinstance(task, BarrierTask)
                and (set(graph.predecessors(task)) | set(graph.successors(task))) & self.selected_tasks
            ):
                yield task

    def get_parent_tasks(self, task: Task) -> Iterable[Task]:
        """
        Returns tasks that must be executed before `task`.
        """
        if task not in self.selected_tasks and not isinstance(task, BarrierTask):
            return

        for parent_task, _ in self.flow.explicit_graph.in_edges(task):
            if parent_task in self.selected_tasks or isinstance(parent_task, BarrierTask):
                yield parent_task

    def visualize(self, result: Result | None = None, visualization_tag: str | None = None):
        dot = self.visualize_pydot(result, visualization_tag)
        _display_pydot(dot)
        return dot

    def visualize_url(self, result: Result | None = None, visualization_tag: str | None = None) -> str:
        if pydot.Dot is None:
            raise RuntimeError("please install pydot to receive a visualization URL")
        dot = self.visualize_pydot(result, visualization_tag)
        return _pydot_url(dot, result.config_context if result else None)

    def visualize_pydot(self, result: Result | None = None, visualization_tag: str | None = None) -> pydot.Dot:
        from pydiverse.pipedag.core import GroupNode, Stage, Task

        flow = self.flow
        # Some group nodes are explicitly wired in the flow at declaration time, but
        # it is also possible to add group nodes via configuration. We must not modify
        # any objects in the flow to make them effective. Thus we create dictionaries
        # on the side to look up config created group nodes.
        task_group_nodes, stage_group_nodes, style_tags = _get_config_group_nodes(flow, result, visualization_tag)

        graph_boxes = set()  # type: Set[Stage | GroupNode]
        graph_nodes = set()  # type: Set[Task | GroupNode]
        graph_edges = set()  # type: Set[Tuple[Task | GroupNode, Task | GroupNode]]
        relevant_group_nodes = set()

        default_style = VisualizationStyle()

        def get_group_node_style(obj: GroupNode):
            return obj.style if obj.style else style_tags.get(obj.style_tag, default_style)

        def get_task_group_node(task: Task):
            return task_group_nodes.get(task, task.group_node if hasattr(task, "group_node") else None)

        def get_stage_outer_group_node(stage: Stage | GroupNode):
            return stage_group_nodes.get(stage, stage.outer_group_node)

        # add tasks and stages to graph and collect group nodes
        for task in self.get_tasks():
            if hasattr(task, "_visualize_hidden") and task._visualize_hidden:
                continue
            graph_boxes.add(task.stage)
            graph_nodes.add(task)
            if group_node := get_task_group_node(task):
                relevant_group_nodes.add(group_node)

            for input_task in task.input_tasks.values():
                graph_nodes.add(input_task)
                graph_boxes.add(input_task.stage)
                if group_node := get_task_group_node(input_task):
                    relevant_group_nodes.add(group_node)

        # add parent stages to graph
        for stage in graph_boxes.copy():
            if stage.outer_stage:
                graph_boxes.add(stage.outer_stage)

        # collect parent group nodes
        for stage in graph_boxes:
            if group_node := get_stage_outer_group_node(stage):
                relevant_group_nodes.add(group_node)

        # get group nodes including selected tasks recursively
        selected_group_nodes = set()
        for group_node in relevant_group_nodes:
            obj = group_node
            stage_tasks = {t for s in obj.stages for t in s.tasks}
            if (obj.tasks | stage_tasks) & self.selected_tasks:
                while True:
                    selected_group_nodes.add(obj)
                    obj = obj.outer_group_node
                    if obj is None:
                        break

        # add parent group nodes to graph
        for node in relevant_group_nodes.copy():
            parent = node.outer_group_node
            while parent:
                relevant_group_nodes.add(parent)
                parent = parent.outer_group_node

        # hide all tasks within hidden group nodes
        for task in graph_nodes.copy():
            if group_node := get_task_group_node(task):
                if group_node.is_content_hidden(get_group_node_style):
                    graph_nodes.remove(task)

        # hide all stages within hidden group nodes
        for stage in graph_boxes.copy():
            if group_node := get_stage_outer_group_node(stage):
                if group_node.is_content_hidden(get_group_node_style):
                    graph_boxes.remove(stage)

        barrier_tasks = set()
        # add group nodes to graph and potentially hide stages and tasks
        for group_node in relevant_group_nodes:
            style = get_group_node_style(group_node)
            if not group_node.outer_group_node or not group_node.outer_group_node.is_content_hidden(
                get_group_node_style
            ):
                if style.hide_box:
                    if group_node.entry_barrier_task:
                        # in-degree > 0 was already verified when creating
                        # entry_barrier_task
                        graph_nodes.add(group_node.entry_barrier_task)
                        barrier_tasks.add(group_node.entry_barrier_task)
                    if (
                        group_node.exit_barrier_task
                        and flow.explicit_graph.out_degree(group_node.exit_barrier_task) > 0
                    ):
                        graph_nodes.add(group_node.exit_barrier_task)
                        barrier_tasks.add(group_node.exit_barrier_task)
                else:
                    if group_node.box_like_stage(style):
                        graph_boxes.add(group_node)
                    else:
                        graph_nodes.add(group_node)

        # add nodes and edges to graph
        graph = nx.DiGraph()

        def get_node(task: Task) -> Task | GroupNode:
            if group_node := get_task_group_node(task):
                if group_node.is_content_hidden(get_group_node_style):
                    obj = group_node
                    while obj.outer_group_node and obj.outer_group_node.is_content_hidden(get_group_node_style):
                        obj = obj.outer_group_node
                    return obj
            return task

        def add_edges(input_tasks: dict[int, Task] | Iterable[Task], target: Task | GroupNode):
            if isinstance(input_tasks, dict):
                input_tasks = input_tasks.values()
            for input_task in input_tasks:
                input_node = get_node(input_task)
                if input_node in graph_nodes:
                    graph_edges.add((input_node, target))

        def add_group_node_edges(group_node: GroupNode):
            if group_node.entry_barrier_task:
                task = group_node.entry_barrier_task
                for prev_task in flow.explicit_graph.predecessors(task):
                    prev_node = get_node(prev_task)
                    if prev_node in graph_nodes:
                        graph_edges.add((prev_node, group_node))
            if group_node.exit_barrier_task:
                task = group_node.exit_barrier_task
                for next_task in flow.explicit_graph.successors(task):
                    next_node = get_node(next_task)
                    if next_node in graph_nodes:
                        graph_edges.add((group_node, next_node))

        for node in graph_nodes:
            if node in barrier_tasks:
                for prev_task in flow.explicit_graph.predecessors(node):
                    prev_node = get_node(prev_task)
                    if prev_node in graph_nodes:
                        graph_edges.add((prev_node, node))
                for next_task in flow.explicit_graph.successors(node):
                    next_node = get_node(next_task)
                    if next_node in graph_nodes:
                        graph_edges.add((node, next_node))
            elif isinstance(node, GroupNode):
                group_node = node
                content_hidden = group_node.is_content_hidden(get_group_node_style)
                for subtask in group_node.tasks:
                    target = group_node if content_hidden else subtask
                    if subtask in self.selected_tasks:
                        add_edges(set(subtask.input_tasks.values()) - group_node.tasks, target)
                add_group_node_edges(group_node)
            else:
                add_edges(node.input_tasks, node)

        for group_node in [b for b in graph_boxes if isinstance(b, GroupNode)]:
            content_hidden = group_node.is_content_hidden(get_group_node_style)
            for subtask in group_node.tasks:
                target = group_node if content_hidden else subtask
                if subtask in self.selected_tasks:
                    add_edges(subtask.input_tasks, target)
            if not get_group_node_style(group_node).hide_box:
                add_group_node_edges(group_node)

        # canonically sort boxes, nodes, and edges
        def stage_sort_key(s):
            return (
                float(s.id)
                if isinstance(s, Stage)
                else stage_sort_key(s.outer_group_node) + 0.01
                if s.outer_group_node
                else s.outer_stage.id + 0.1
                if s.outer_stage
                else min(s2.id for s2 in s.stages) - 0.1
            ) + (0.001 * s.id if isinstance(s, GroupNode) else 0)

        def task_sort_key(t):
            return (
                t.id
                if isinstance(t, Task)
                else min(t.id for t in t.tasks) - 0.1
                if t.tasks
                else min(stage_sort_key(s) for s in t.stages) - 0.01
            )

        def edge_sort_key(e):
            return (task_sort_key(e[0]), task_sort_key(e[1]))

        sorted_boxes = sorted(graph_boxes, key=stage_sort_key)
        sorted_nodes = sorted(graph_nodes, key=task_sort_key)
        sorted_edges = sorted(graph_edges, key=edge_sort_key)

        # fill graph
        for node in sorted_nodes:
            graph.add_node(node)
        for node1, node2 in sorted_edges:
            graph.add_edge(node1, node2)

        # prepare styles
        stage_style = {}
        task_style = _generate_task_style(graph_nodes, result)
        group_node_style = _generate_group_node_style(relevant_group_nodes, result, get_group_node_style)
        edge_style = {}

        for stage in graph_boxes:
            if stage not in self.selected_stages:
                stage_style[stage] = {
                    "style": '"dashed"',
                    "bgcolor": "#0000000A",
                    "color": "#909090",
                    "fontcolor": "#909090",
                }

        for node in graph_nodes:
            if node not in self.selected_tasks and node not in barrier_tasks:
                task_style[node] = {
                    "style": '"filled,dashed"',
                    "fillcolor": "#FFFFFF80",
                    "color": "#808080",
                    "fontcolor": "#909090",
                }

        for group_node in relevant_group_nodes:
            style = get_group_node_style(group_node)
            if not style.hide_box and group_node in group_node_style:
                if group_node.box_like_stage(style):
                    stage_style[group_node] = group_node_style[group_node]
                    if group_node_style[group_node].get("style") == "filled":
                        # included stages shall not inherit color from group node
                        for stage in group_node.stages:
                            stage_style[stage] = {
                                "style": "filled",
                                "fillcolor": "#dddddd",
                            }
                else:
                    task_style[group_node] = group_node_style[group_node]
            if group_node not in selected_group_nodes and not style.hide_box:
                if group_node.box_like_stage(style):
                    stage_style[group_node] = {
                        "style": '"dashed"',
                        "bgcolor": "#0000000A",
                        "color": "#909090",
                        "fontcolor": "#909090",
                    }
                else:
                    task_style[group_node] = {
                        "style": '"filled,dashed"',
                        "fillcolor": "#FFFFFF80",
                        "color": "#808080",
                        "fontcolor": "#909090",
                    }

        for edge in graph.edges:
            if (
                edge[0] not in self.selected_tasks
                and edge[0] not in barrier_tasks
                and (not isinstance(edge[0], GroupNode) or all(t not in self.selected_tasks for t in edge[0].tasks))
            ):
                edge_style[edge] = {
                    "style": "dashed",
                    "color": "#909090",
                }

        return _build_pydot(
            stages=list(sorted_boxes),
            tasks=list(sorted_nodes),
            get_task_group_node=get_task_group_node,
            get_stage_outer_group_node=get_stage_outer_group_node,
            get_group_node_style=get_group_node_style,
            graph=graph,
            stage_style=stage_style,
            task_style=task_style,
            edge_style=edge_style,
        )


# Visualization Helper


def _generate_task_style(tasks: Iterable[Task], result: Result | None) -> dict:
    task_style = {}
    if result:
        for task in tasks:
            final_state = result.task_states.get(task, FinalTaskState.UNKNOWN)
            if final_state == FinalTaskState.COMPLETED:
                task_style[task] = {"fillcolor": "#adef9b"}
            elif final_state == FinalTaskState.CACHE_VALID:
                task_style[task] = {"fillcolor": "#e8ffc6"}
            elif final_state == FinalTaskState.FAILED:
                task_style[task] = {"fillcolor": "#ff453a"}
            elif final_state == FinalTaskState.UNKNOWN:
                task_style[task] = {"fillcolor": "#ffffff"}
    return task_style


def _generate_group_node_style(
    group_nodes: Iterable[GroupNode],
    result: Result | None,
    get_style: Callable[[GroupNode], VisualizationStyle],
) -> dict:
    group_node_style = {}
    if result:
        for group_node in group_nodes:
            style = get_style(group_node)
            final_states = {result.task_states.get(task, FinalTaskState.UNKNOWN) for task in group_node.tasks}
            if not style.hide_box:
                if style.box_color_always:
                    group_node_style[group_node] = {
                        "style": "filled",
                        "fillcolor": style.box_color_always,
                    }
                elif style.hide_content:
                    if FinalTaskState.FAILED in final_states:
                        fill_color = style.box_color_any_failure or "#ff453a"
                        group_node_style[group_node] = {
                            "style": "filled",
                            "fillcolor": fill_color,
                        }
                    elif final_states == {FinalTaskState.UNKNOWN}:
                        fill_color = style.box_color_all_skipped or "#ffffff"
                        group_node_style[group_node] = {
                            "style": "filled",
                            "fillcolor": fill_color,
                        }
                    elif FinalTaskState.CACHE_VALID not in final_states:
                        fill_color = style.box_color_none_cache_valid or "#adef9b"
                        group_node_style[group_node] = {
                            "style": "filled",
                            "fillcolor": fill_color,
                        }
                    elif final_states == {FinalTaskState.CACHE_VALID}:
                        fill_color = style.box_color_all_cache_valid or "#e8ffc6"
                        group_node_style[group_node] = {
                            "style": "filled",
                            "fillcolor": fill_color,
                        }
                    else:
                        fill_color = style.box_color_any_cache_valid or "#caffaf"
                        group_node_style[group_node] = {
                            "style": "filled",
                            "fillcolor": fill_color,
                        }
                else:
                    group_node_style[group_node] = {"style": "solid"}
    return group_node_style


def _build_pydot(
    stages: list[Stage | GroupNode],
    tasks: list[Task | GroupNode],
    get_task_group_node: Callable[[Task], GroupNode],
    get_stage_outer_group_node: Callable[[Stage], GroupNode],
    get_group_node_style: Callable[[GroupNode], VisualizationStyle],
    graph: nx.Graph,
    stage_style: dict[Stage | GroupNode, dict] | None,
    task_style: dict[Task | GroupNode, dict] | None,
    edge_style: dict[tuple[Task | GroupNode, Task | GroupNode], dict] | None,
) -> pydot.Dot:
    """
    Build a pydot graph from a graph of tasks and stages.

    :param stages:
        List of boxes around tasks be it stages or group nodes
    :param tasks:
        List of nodes in the graph be it tasks or group nodes
    :param get_task_group_node:
        Function to get the group node of a task. We can't use task.group_node directly
        because group nodes added by configuration should not modify the flow.
    :param get_stage_outer_group_node:
        Function to get the outer group node of a stage. We can't use
        stage.outer_group_node directly because group nodes added by configuration
        should not modify the flow.
    :param graph:
        The graph representing edges between tasks and stages
    :param stage_style:
        Style of boxes be it stages or group nodes
    :param task_style:
        Style of nodes in the graph be it tasks or group nodes
    :param edge_style:
        Style of edges in the graph
    :return:
        A pydot graph
    """
    from pydiverse.pipedag.core import GroupNode

    if stage_style is None:
        stage_style = {}
    if task_style is None:
        task_style = {}
    if edge_style is None:
        edge_style = {}

    if pydot.Dot is None:
        raise RuntimeError("Please install pydot to visualize the flow execution.")

    dot = pydot.Dot(compound="true")
    subgraphs: dict[Stage | GroupNode, pydot.Subgraph] = {}
    nodes: dict[Task | GroupNode, pydot.Node] = {}

    for stage in stages:
        style = dict(
            style="solid",
            bgcolor="#00000020",
            color="black",
            fontcolor="black",
        ) | (stage_style.get(stage, {}))

        if isinstance(stage, GroupNode):
            group_node = stage
            label = group_node.label or ""
            if get_group_node_style(group_node).hide_label:
                label = ""
            s = pydot.Cluster(
                f"n_{base64.b64encode(random.randbytes(8)).decode('ascii')}",
                label=label,
                **style,
            )
        else:
            s = pydot.Cluster(f"s_{stage.name}", label=stage.name, **style)
        subgraphs[stage] = s

    for stage in stages:
        s = subgraphs[stage]
        stage_outer_group_node = get_stage_outer_group_node(stage)
        if stage.outer_stage in stages and get_stage_outer_group_node(stage.outer_stage) is stage_outer_group_node:
            subgraphs[stage.outer_stage].add_subgraph(s)
        elif stage_outer_group_node in stages and (
            not stage.outer_stage or stage.outer_stage not in stage_outer_group_node.stages
        ):
            subgraphs[stage_outer_group_node].add_subgraph(s)
        elif stage.outer_stage in stages:
            subgraphs[stage.outer_stage].add_subgraph(s)
        else:
            dot.add_subgraph(s)

    for task in tasks:
        style = dict(
            fillcolor="#FFFFFF",
            style='"filled"',
        ) | (task_style.get(task, {}))

        if isinstance(task, GroupNode):
            group_node = task
            label = group_node.label or "|".join(
                sorted([t.name for t in group_node.stages]) + sorted([t.name for t in group_node.tasks])
            )
            if get_group_node_style(group_node).hide_label:
                label = ""
            node = pydot.Node(
                base64.b64encode(random.randbytes(8)).decode("ascii"),
                label=label,
                **style,
            )
            if group_node.outer_group_node in stages and (
                not group_node.outer_stage or group_node.outer_stage not in group_node.outer_group_node.stages
            ):
                subgraphs[group_node.outer_group_node].add_node(node)
            elif group_node.outer_stage in stages:
                subgraphs[group_node.outer_stage].add_node(node)
            else:
                dot.add_node(node)
        else:
            node = pydot.Node(task.id, label=task.name, **style)
            task_group_node = get_task_group_node(task)
            if (
                hasattr(task, "group_node")
                and task_group_node
                and task_group_node in stages
                and task.stage not in task_group_node.stages
            ):
                subgraphs[task_group_node].add_node(node)
            else:
                subgraphs[task.stage].add_node(node)
        nodes[task] = node

    for nx_edge in graph.edges:
        style = edge_style.get(nx_edge, {})
        sender = nodes[nx_edge[0]] if nx_edge[0] in nodes else subgraphs[nx_edge[0]]
        receiver = nodes[nx_edge[1]] if nx_edge[1] in nodes else subgraphs[nx_edge[1]]
        sender_task = sender if nx_edge[0] in nodes else nodes[list(nx_edge[0].tasks)[0]]
        receiver_task = receiver if nx_edge[1] in nodes else nodes[list(nx_edge[1].tasks)[0]]
        edge = pydot.Edge(
            sender_task,
            receiver_task,
            ltail=sender.get_name(),
            lhead=receiver.get_name(),
            **style,
        )
        dot.add_edge(edge)

    return dot


def _display_pydot(dot: pydot.Dot):
    """Display a pydot.Dot graph

    - Either as svg in ipython
    - Or as PDF in default pdf viewer
    """

    try:
        from IPython import get_ipython
        from IPython.display import SVG, display

        ipython = get_ipython()
        if ipython is None or ipython.config.get("IPKernelApp") is None:
            raise RuntimeError("Either IPython isn't running or SVG aren't supported.")
        display(SVG(dot.create_svg()))  # type: ignore
    except (ImportError, RuntimeError):
        import tempfile
        import webbrowser

        f = tempfile.NamedTemporaryFile(suffix=".pdf")
        f.write(dot.create_pdf())  # type: ignore
        webbrowser.open_new("file://" + f.name)

        # Keep alive to prevent immediate deletion of file
        globals()["__pipedag_tmp_file_reference__"] = f


def _pydot_url(dot: pydot.Dot, config: ConfigContext | None = None) -> str:
    import base64
    import zlib

    if config is None:
        try:
            config = ConfigContext.get()
        except LookupError:
            config = PipedagConfig.default.get()

    kroki_url = config.kroki_url
    if config.disable_kroki or kroki_url is None:
        kroki_url = "<disable_kroki=True>"

    query_data = zlib.compress(str(dot).encode("utf-8"), 9)
    query = base64.urlsafe_b64encode(query_data).decode("ascii")

    return f"{kroki_url}/graphviz/svg/{query}"


def _get_config_group_nodes(flow: Flow, result: Result | None, visualization_tag: str | None):
    from pydiverse.pipedag.core import GroupNode

    logger = flow.logger
    if result:
        config_context = result.config_context
    else:
        try:
            config_context = ConfigContext.get()
        except LookupError:
            config_context = PipedagConfig.default.get(flow=flow.name)

    group_nodes, task_group_nodes, stage_group_nodes, style_tags = {}, {}, {}, {}
    if visualization_tag is None:
        visualization_tag = "default"
    if visualization_tag in config_context.visualization:
        visualization = config_context.visualization[visualization_tag]
        style_tags = visualization.styles or {}
        if visualization.group_nodes:
            for node_tag, node_config in visualization.group_nodes.items():
                style = visualization.styles.get(node_config.style_tag) if visualization.styles else None
                group_node = GroupNode(node_config.label, style, style_tag=node_config.style_tag)
                group_node.id = len(group_nodes) + len(flow.group_nodes)
                group_nodes[node_tag] = group_node
                for stage_name in node_config.stages or []:
                    if stage_name in flow.stages:
                        stage = flow.stages[stage_name]
                        group_node.add_stage(stage)
                        stage_group_nodes[stage] = group_node
                    else:
                        logger.error(
                            f"Stage {stage_name} in config based visualization "
                            f"group node {visualization_tag}.{node_tag} not found "
                            f"in flow."
                        )
                for stage in flow.stages.values():
                    if stage.group_node_tag == node_tag:
                        group_node.add_stage(stage)
                        stage_group_nodes[stage] = group_node

                for task_name in node_config.tasks or []:
                    found = False
                    for task in flow.tasks:
                        if task.name == task_name:
                            if found:
                                raise ValueError(
                                    f"Task {task_name} in config based "
                                    f"visualization group node "
                                    f"{visualization_tag}.{node_tag} appears "
                                    f"multiple times in flow."
                                )
                            found = True
                            if other_group_node := stage_group_nodes.get(task.stage):
                                if other_group_node is group_node:
                                    logger.info(
                                        f"Ignoring task {task_name} in config based"
                                        f" visualization group node "
                                        f"{visualization_tag}.{node_tag} because "
                                        f"its whole stage is included."
                                    )
                                    continue
                                if group_node.outer_group_node and group_node.outer_group_node is not other_group_node:
                                    raise ValueError(
                                        f"Configured group node {node_tag} includes"
                                        f" tasks from multiple outer group nodes "
                                        f"which is not allowed: "
                                        f"{group_node.outer_group_node}, "
                                        f"{other_group_node}"
                                    )
                                group_node.outer_group_node = other_group_node
                                if (
                                    group_node.outer_stage
                                    and group_node.outer_stage not in group_node.stages
                                    and group_node.outer_stage is not other_group_node.outer_stage
                                ):
                                    raise ValueError(
                                        f"Configured group node {node_tag} includes"
                                        f" tasks from multiple stages which is not"
                                        f" allowed: "
                                        f"{group_node.outer_stage}, "
                                        f"{other_group_node.outer_stage}, "
                                        f"included_stages={group_node.stages}"
                                    )
                                group_node.outer_stage = other_group_node.outer_stage

                            if group_node.outer_stage and group_node.outer_stage is not task.stage:
                                raise ValueError(
                                    f"Configured group node {node_tag} includes "
                                    f"tasks from multiple stages which is not "
                                    f"allowed: "
                                    f"{group_node.outer_stage}, {task.stage}"
                                )
                            group_node.outer_stage = task.stage
                            group_node.add_task(task)
                            if task.group_node:
                                logger.info(
                                    f"Adding task {task_name} to config based "
                                    f"visualization group node "
                                    f"{visualization_tag}.{node_tag}. The flow "
                                    f"based membership in {task.group_node} will "
                                    f"not take any effect (empty groups will not be"
                                    f" displayed)."
                                )
                            if task in task_group_nodes:
                                raise ValueError(
                                    f"Task {task_name} in config based "
                                    f"visualization group node "
                                    f"{visualization_tag}.{node_tag} appears"
                                    f" in multiple configured group nodes: "
                                    f"{task_group_nodes[task]}, {group_node}."
                                )
                            task_group_nodes[task] = group_node
                    if not found:
                        logger.error(
                            f"Task {task_name} in config based visualization "
                            f"group node {visualization_tag}.{node_tag} not found "
                            f"in flow."
                        )

                for task in flow.tasks:
                    if hasattr(task, "group_node_tag") and task.group_node_tag == node_tag:
                        task_group_nodes[task] = group_node
                        group_node.add_task(task)

    elif visualization_tag != "default":
        logger.error(f"Visualization tag {visualization_tag} not found in config.")
    else:
        logger.info("No visualization customization found in config.")

    def find_outer_group_node(stage: Stage, stages: list[Stage]):
        if stage in stage_group_nodes:
            return stage_group_nodes[stage], stages + [stage]
        if stage.outer_stage:
            return find_outer_group_node(stage.outer_stage, stages + [stage])
        if stage.outer_group_node:
            return stage.outer_group_node, stages + [stage]
        return None, []

    def update_outer_group_node(stage: Stage):
        stage_outer_group_node, stages = find_outer_group_node(stage, [])
        if stage_outer_group_node:
            for stage in stages:
                stage_group_nodes[stage] = stage_outer_group_node
        return stage_outer_group_node

    # link tasks and stages to next outer group nodes
    for task in [t for t in flow.tasks if t not in task_group_nodes]:
        if hasattr(task, "group_node") and not task.group_node:
            if group_node := update_outer_group_node(task.stage):
                task_group_nodes[task] = group_node

    # link group nodes to outer group nodes and stages
    for group_node in group_nodes.values():
        outer_node = group_node.outer_group_node
        for task in group_node.tasks:
            outer_node = outer_node or update_outer_group_node(task.stage)
        for stage in group_node.stages:
            stage_outer_group_node = update_outer_group_node(stage)
            if stage_outer_group_node is not group_node:
                outer_node = outer_node or stage_outer_group_node
        group_node.outer_group_node = outer_node

    return task_group_nodes, stage_group_nodes, style_tags



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import sqlalchemy as sa

from pydiverse.pipedag import Flow, Stage, materialize
from pydiverse.pipedag.context import StageLockContext

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import (
    ALL_INSTANCES,
    ORCHESTRATION_INSTANCES,
    skip_instances,
    with_instances,
)
from tests.util import tasks_library as m
from tests.util.tasks_library import simple_dataframe

pytestmark = [with_instances(ALL_INSTANCES, ORCHESTRATION_INSTANCES)]


def test_unicode(unicode_str="äöüßéç"):
    @materialize(lazy=True, input_type=sa.Table)
    def unicode(src):
        return sa.select(sa.literal(unicode_str).label("a")).select_from(src).limit(1)

    with Flow("flow") as f:
        with Stage("stage"):
            dummy_source = simple_dataframe()
            x = unicode(dummy_source)
            x2 = m.noop(x)
            x3 = m.noop_lazy(x2)
            m.assert_table_equal(x, x2)
            m.assert_table_equal(x, x3)

    with StageLockContext():
        result = f.run()
        assert result.successful
        assert result.get(x3, as_type=pd.DataFrame)["a"][0] == unicode_str


@skip_instances("mssql", "mssql_pytsql", "mssql_nobulk", "mssql_noaodbc", "mssql_columnstore")
def test_unicode_beyond_mssql():
    test_unicode("λ")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from tests.util.pytest_raises import swallowing_raises
from tests.util.sql import select_as

__all__ = [
    "swallowing_raises",
    "select_as",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

import sqlalchemy as sa

from pydiverse.pipedag import Flow, Stage, materialize
from pydiverse.pipedag.container import RawSql
from pydiverse.pipedag.context import ConfigContext, StageLockContext
from tests.fixtures.instances import with_instances

"""
Attention:
Wrapping Raw SQL statements should always be just the first step of pipedag adoption.
Ideally, the next step is to extract individual transformations (SELECT statements) so
they can be gradually converted from text SQL to programmatically created SQL (python).
"""


@materialize(input_type=sa.Table, lazy=True)
def tsql(
    name: str,
    script_directory: Path,
    *,
    out_stage: Stage | None = None,
    in_sql=None,
    helper_sql=None,
    depend=None,
):
    _ = depend  # only relevant for adding additional task dependency
    script_path = script_directory / name
    sql = Path(script_path).read_text(encoding="utf-8")
    sql = raw_sql_bind_schema(sql, "out_", out_stage, transaction=True)
    sql = raw_sql_bind_schema(sql, "in_", in_sql)
    sql = raw_sql_bind_schema(sql, "helper_", helper_sql)
    return RawSql(sql, Path(script_path).name)


def raw_sql_bind_schema(sql, prefix: str, stage: Stage | RawSql | None, *, transaction=False):
    if isinstance(stage, RawSql):
        stage = stage.stage
    config = ConfigContext.get()
    store = config.store.table_store
    if stage is not None:
        stage_name = stage.transaction_name if transaction else stage.name
        schema = store.get_schema(stage_name).get()
        sql = sql.replace("{{%sschema}}" % prefix, schema)  # noqa: UP031
    return sql


@with_instances("mssql")
def test_raw_sql():
    instance_name = ConfigContext.get().instance_name
    parent_dir = Path(__file__).parent / "raw_sql_scripts" / instance_name

    with Flow() as flow:
        with Stage("helper") as out_stage:
            helper = tsql("create_db_helpers.sql", parent_dir, out_stage=out_stage)
        with Stage("raw") as out_stage:
            _dir = parent_dir / "raw"
            raw = tsql("raw_views.sql", _dir, out_stage=out_stage, helper_sql=helper)
        with Stage("prep") as prep_stage:
            _dir = parent_dir / "prep"
            prep = tsql("entity_checks.sql", _dir, in_sql=raw, out_stage=prep_stage, depend=raw)
            prep = tsql("more_tables.sql", _dir, in_sql=raw, out_stage=prep_stage, depend=prep)
            _ = prep

    # on a fresh database, this will create indexes with Raw-SQL
    _run_and_check(flow, prep_stage)
    # make sure cached execution creates the same indexes
    _run_and_check(flow, prep_stage)


def _run_and_check(flow, prep_stage):
    config_ctx = ConfigContext.get()

    with StageLockContext():
        flow_result = flow.run()
        assert flow_result.successful

        schema = config_ctx.store.table_store.get_schema(prep_stage.name).get()
        inspector = sa.inspect(config_ctx.store.table_store.engine)

        # these constraints might be a bit too harsh in case this test is extended
        # to more databases
        assert set(inspector.get_table_names(schema=schema)) == {
            "raw01A",
            "table01",
            "special_chars",
            "special_chars2",
            "special_chars_join",
        }

        pk = inspector.get_pk_constraint("raw01A", schema=schema)
        assert pk["constrained_columns"] == ["entity", "start_date"]
        assert pk["name"].startswith("PK__raw01A__")

        pk = inspector.get_pk_constraint("table01", schema=schema)
        assert pk["constrained_columns"] == ["entity", "reason"]
        assert pk["name"].startswith("PK__table01__")
        assert len(inspector.get_indexes("table01", schema=schema)) == 0

        indexes = inspector.get_indexes("raw01A", schema=schema)
        assert indexes[0]["name"] == "raw_start_date"
        assert indexes[0]["column_names"] == ["start_date"]
        assert indexes[1]["name"] == "raw_start_date_end_date"
        assert indexes[1]["column_names"] == ["end_date", "start_date"]

        with config_ctx.store.table_store.engine.connect() as conn:
            sql = f"SELECT string_col FROM [{schema}].[special_chars_join]"
            str_val = conn.execute(sa.text(sql)).fetchone()[0]
            assert str_val == "äöüßéç"



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import contextlib
import datetime as dt
import re
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from types import GenericAlias
from typing import get_args, get_origin

import pytest
import structlog

from pydiverse.pipedag import (
    Blob,
    ConfigContext,
    Flow,
    PipedagConfig,
    RawSql,
    Stage,
    Table,
)
from pydiverse.pipedag.context import RunContext, RunContextServer, default_config_dict
from pydiverse.pipedag.context.context import test_store_config_dict
from pydiverse.pipedag.context.trace_hook import TraceHook
from pydiverse.pipedag.util.json import PipedagJSONDecoder, PipedagJSONEncoder


def deep_cmp(a, b):
    if a != b:
        logger = structlog.get_logger("test_json.deep_cmp")
        logger.info("Found difference", a=a, b=b)
    if isinstance(a, str):
        return a == b
    if isinstance(a, dict):
        return all([deep_cmp(v, b[k]) for k, v in a.items()])
    if isinstance(a, GenericAlias):
        base_cmp = get_origin(a) == get_origin(b)
        return base_cmp and all([deep_cmp(v, w) for v, w in zip(get_args(a), get_args(b))])
    if isinstance(a, Iterable):
        return all([deep_cmp(v, w) for v, w in zip(a, b)])
    if hasattr(a, "__dict__"):

        def fields(x):
            return {k: v for k, v in x.__dict__.items() if not k.startswith("_") and k != "logger"}

        return deep_cmp(fields(a), fields(b))
    return a == b


def check(x, expected_result=None, regex: str | None = None, regex_replace: str | None = None):
    expected_result = expected_result or x
    json_encoder = PipedagJSONEncoder()
    json_decoder = PipedagJSONDecoder()
    y = json_encoder.encode(x)
    if regex:
        regex_replace = regex_replace or ""
        y = re.sub(regex, regex_replace, y, flags=re.MULTILINE)
    z = json_decoder.decode(y)
    assert deep_cmp(expected_result, z)


def test_json_coder_primitive():
    x = {
        "a": 1,
        "b": 2.0,
        "c": "3",
        "d": True,
        "e": None,
        "f": "45",
        "g": "",
    }
    check(x)


def set_stage(x, s: Stage):
    x.stage = s
    return x


def set_cache_key(x, k: str):
    x.cache_key = k
    return x


def _cfg_ctx():
    return ConfigContext.new(default_config_dict | test_store_config_dict, "x", "y", "z")


@contextlib.contextmanager
def _with_dummy_context(*, stage_name: str = "x"):
    # Setup flow so Stage("h") exists
    flow = Flow()
    flow.stages[stage_name] = Stage(stage_name)
    with _cfg_ctx(), RunContext(RunContextServer(flow.get_subflow(), TraceHook())):
        yield


def test_json_coder_table():
    class C:
        t: dt.datetime

    x = {
        "a": Table(),
        "b": Table(name="b"),
        "c": Table(primary_key=["c"]),
        "d": Table(annotation=C.__annotations__["t"]),
        "e": Table(annotation=C.__annotations__),
        "f": Table(annotation=list[int]),
        "g": Table(annotation=dict[str, dt.datetime]),
        "h": set_stage(Table(), Stage("h")),
        "i": set_cache_key(Table(), "i"),
    }
    with _with_dummy_context(stage_name="h"):
        check(x)


def test_json_coder_table_fallback():
    class C:
        t: dt.datetime

    x = {
        "a": Table(),
        "d": Table(name="d", annotation=C.__annotations__["t"]),
        "e": Table(annotation=C.__annotations__),
    }
    expected_result = {
        "a": Table(),
        "d": Table(name="d"),
        "e": Table(),
    }
    check(
        x,
        expected_result,
        regex=r'"annotation":([^,]*{[^}]*(}(?!,)[^}]*)?},|[^,{}]*,)',
        regex_replace="",
    )


def test_json_coder_table_fallback2():
    check(
        Table(annotation=list[int]),
        Table(),
        regex=r'"annotation":[^}]*}[^}]*}[^}]*},',
        regex_replace="",
    )
    check(
        Table(annotation=dict[str, dt.datetime]),
        Table(),
        regex=r'"annotation":[^}]*}[^}]*}[^}]*}[^}]*},',
        regex_replace="",
    )


def test_json_coder_table_obj_remove():
    x = {
        "a": Table(1),
    }
    with pytest.raises(AssertionError):
        check(x)


def test_json_coder_blob():
    class C:
        t: dt.datetime

    x = {
        "a": Blob(),
        "b": Blob(name="b"),
        "c": set_stage(Blob(), Stage("c")),
        "d": set_cache_key(Blob(), "dd"),
    }
    with _with_dummy_context(stage_name="c"):
        check(x)


def test_json_coder_blob_obj_remove():
    x = {
        "a": Blob(1),
    }
    with pytest.raises(AssertionError):
        check(x)


def test_json_coder_raw_sql():
    class C:
        t: dt.datetime

    x = {
        "a": RawSql(),
        "b": RawSql(name="b"),
        "c": set_stage(RawSql(), Stage("c")),
        "d": set_cache_key(RawSql(), "dd"),
        "e": RawSql(separator=" GO "),
        "f": RawSql("select 1"),
    }
    with _with_dummy_context(stage_name="c"):
        check(x)


def test_json_coder_classes():
    x = {
        "a": Stage("a"),
        "c": PipedagConfig.default.get(),
        "d": _cfg_ctx(),
        "e": Path(),
        "f": dt.datetime(1, 2, 3),
        "g": dt.datetime(1, 2, 3, 4, 5),
        "h": dt.datetime(1, 2, 3, 4, 5, 6),
        "i": dt.date(1, 2, 3),
    }
    with _with_dummy_context(stage_name="a"):
        check(x)


def test_json_coder_pipedag_config_only_input():
    # PipedagConfig is not supposed to be deserialized well. It is only needed
    # as input for getting another instance ConfigContext
    x = {
        "b": PipedagConfig.default,
    }
    with pytest.raises(KeyError, match="table_store_connection"):
        check(x)


@dataclass
class C:
    m: int
    n: str
    o: list[int]


@dataclass
class D:
    m: Table
    n: Table[dt.datetime]
    o: Table[int]


def test_json_coder_dataclass():
    x = {
        "a": C(1, "2", [3, 4, 5]),
        "b": D(Table(), Table(), Table()),
        "c": D(Table(), Table(annotation=dt.datetime), Table(annotation=int)),
    }
    check(x)


def test_json_coder_types():
    x = {
        "a": C,
        "b": D,
        "c": int,
        "d": dt.datetime,
        "e": str,
    }
    check(x)


def test_json_coder_fail_local_dataclass():
    @dataclass
    class C:
        m: int
        n: str
        o: list[int]

    x = {
        "a": C(1, "2", [3, 4, 5]),
    }
    with pytest.raises(AttributeError, match="'function' object has no attribute '<locals>'"):
        check(x)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import sqlalchemy as sa
from sqlalchemy import inspect

from pydiverse.pipedag import Blob, RawSql, Table, materialize
from pydiverse.pipedag.backend.table.sql.ddl import MAX_LENGTH_PK
from pydiverse.pipedag.context import TaskContext
from pydiverse.pipedag.debug import materialize_table
from tests.util import select_as

try:
    import polars as pl
except ImportError:
    pl = None


@materialize(input_type=pd.DataFrame, version="1.0")
def noop(x):
    return x


@materialize(input_type=pd.DataFrame, version="1.0")
def noop2(x):
    return x


@materialize(input_type=sa.Table, version="1.0")
def noop_sql(x):
    return x


@materialize(input_type=sa.Table, lazy=True)
def noop_lazy(x):
    return x


@materialize(input_type=pl.DataFrame if pl else "<polars task without polars>")
def noop_polars(x):
    return Table(x)


@materialize(input_type=pl.LazyFrame if pl else "<polars task without polars>")
def noop_lazy_polars(x):
    return Table(x)


@materialize(input_type=pd.DataFrame, version="1.0")
def noop_subtask(x):
    @materialize(input_type=pd.DataFrame, version="1.0")
    def _noop(x):
        return Table(x)

    # This is somewhat crazy behavior to call a task within a task. However, since we
    # support calling tasks outside of `with Flow`, this will work for dataframe tasks.
    return _noop(x)


@materialize(input_type=sa.Table, lazy=True)
def noop_subtask_lazy(x):
    @materialize(input_type=sa.Table, lazy=True)
    def _noop(x):
        return Table(x)

    # This is somewhat crazy behavior to call a task within a task. However, since we
    # support calling tasks outside of `with Flow`, this will work for same type.
    return _noop(x)


@materialize(input_type=pd.DataFrame, version="1.0")
def noop_subtask_fail_input_type(x):
    @materialize(input_type=sa.Table, version="1.0")
    def _noop(x):
        return Table(x)

    # calling task with different input type should fail with good error message
    # because we call subtasks without separate dematerialization
    return _noop(x)


@materialize(nout=2, version="1.0", input_type=pd.DataFrame)
def create_tuple(x, y):
    return x, y


@materialize(version="1.0")
def one():
    return 1


@materialize(lazy=True)
def one_sql_lazy():
    return select_as(1, "x")


@materialize(version="1.0")
def two():
    return 2


@materialize()
def assert_equal(x, y):
    # This function explicitly has no version set to prevent it from getting cached
    assert x == y, f"{x} != {y}"


@materialize(input_type=pd.DataFrame)
def assert_table_equal(x, y, **kwargs):
    # This function explicitly has no version set to prevent it from getting cached
    pd.testing.assert_frame_equal(x, y, **kwargs)


@materialize()
def assert_blob_equal(x, y):
    # This function explicitly has no version set to prevent it from getting cached
    assert x == y


@materialize(version="1.0", input_type=pd.DataFrame)
def take_first(table, as_int=False):
    ret = table["x"][0]
    if as_int:
        # might be needed to avoid returning unserializable numpy.int64
        # we could also consider JSON serializing numpy equivalents of primitive types
        # either conserving them being numpy types or converting them to primitive types
        return int(ret)
    return ret


@materialize(version="1.0")
def simple_dataframe():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    return Table(df)


@materialize(version="1.0")
def simple_dataframe_subtask():
    # This is somewhat crazy behavior to call a task within a task. However, since we
    # support calling tasks outside of `with Flow`, this will work for dataframe tasks.
    return simple_dataframe()


@materialize(version="1.0", nout=2)
def simple_dataframe_uncompressed():
    return (
        Table(
            pd.DataFrame(dict(COMPRESSION=["N"], ROWCOMPMODE=[" "])),
            "simple_df_properties",
        ),
        simple_dataframe(),
    )


@materialize(nout=2)
def simple_dataframe_compressed_one_method():
    return Table(
        pd.DataFrame(dict(COMPRESSION=["R"], ROWCOMPMODE=["S"])),
        "df_compressed_1_properties",
    ), _simple_dataframe_materialization_details("df_compressed_1", "static_compression")


@materialize(nout=2)
def simple_dataframe_compressed_two_methods():
    return Table(
        pd.DataFrame(dict(COMPRESSION=["B"], ROWCOMPMODE=["A"])),
        "df_compressed_2_properties",
    ), _simple_dataframe_materialization_details("df_compressed_2", "adaptive_value_compression")


def _simple_dataframe_materialization_details(name=None, materialization_details=None):
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    return Table(df, name=name, materialization_details=materialization_details)


@materialize(version="1.0")
def simple_dataframe_debug_materialize_no_taint():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    res = Table(df, name="test_table")
    materialize_table(res, flag_task_debug_tainted=False, debug_suffix="debug")
    return res


@materialize(version="1.0")
def simple_dataframe_debug_materialize_twice():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    res = Table(df)
    materialize_table(res, flag_task_debug_tainted=True, debug_suffix="debug")

    df.iloc[3] = [4, "4"]
    res = Table(df)
    materialize_table(res, flag_task_debug_tainted=True, debug_suffix="debug")

    return res


def complex_imperative_materialize(in_: Table):
    @materialize(lazy=True, input_type=sa.Table)
    def _temp_res1(in_: sa.Table):
        query = sa.select("*").select_from(in_)
        return Table(query)

    @materialize(lazy=True, input_type=sa.Table)
    def _temp_res2(in_: sa.Table):
        query1 = sa.select("*").select_from(in_)
        t1 = Table(query1).materialize()

        query2 = sa.select("*").select_from(t1)
        t2 = Table(query2).materialize()

        query3 = sa.select("*").select_from(t2)
        return Table(query3)

    @materialize(lazy=True, input_type=sa.Table)
    def _temp_res3(_temp_res1: sa.Table, _temp_res2: sa.Table):
        _ = _temp_res2
        query = sa.select("*").select_from(_temp_res1)
        return Table(query)

    temp_res1 = _temp_res1(in_)
    temp_res2 = _temp_res2(temp_res1)
    return _temp_res3(temp_res1, temp_res2)


@materialize(version="1.0")
def simple_dataframe_with_pk():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    return Table(df, primary_key="col1")


@materialize(version="1.0")
def simple_dataframe_with_pk2():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    return Table(df, primary_key=["col1", "col2"])


@materialize(version="1.0")
def simple_dataframe_with_index():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    return Table(df, primary_key=["col1"], indexes=[["col2"]])


@materialize(version="1.0")
def simple_dataframe_with_indexes():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    return Table(df, primary_key=["col1"], indexes=[["col2"], ["col2", "col1"]])


@materialize(nout=2, version="1.0")
def simple_dataframes_with_indexes_task():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    return Table(df, "dfA", primary_key=["col1"], indexes=[["col2"], ["col2", "col1"]]), Table(
        df, "dfB", primary_key=["col1"], indexes=[["col2"], ["col2", "col1"]]
    )


def simple_dataframes_with_indexes():
    res, _ = simple_dataframes_with_indexes_task()
    return res


def _get_df_query():
    try:
        unions = [sa.select([sa.literal(i).label("col1"), sa.literal(str(i)).label("col2")]) for i in range(4)]
    except sa.exc.ArgumentError:
        # this works from sqlalchemy 2.0.0 on
        unions = [sa.select(sa.literal(i).label("col1"), sa.literal(str(i)).label("col2")) for i in range(4)]
    return unions[0].union_all(*unions[1:])


@materialize(lazy=True)
def simple_lazy_table():
    query = _get_df_query()
    return Table(query)


@materialize(lazy=True)
def simple_lazy_table_with_pk():
    query = _get_df_query()
    return Table(query, primary_key="col1")


@materialize(lazy=True)
def simple_lazy_table_with_pk2():
    query = _get_df_query()
    return Table(query, primary_key=["col1", "col2"])


@materialize(lazy=True)
def simple_lazy_table_with_index():
    query = _get_df_query()
    return Table(query, primary_key=["col1"], indexes=[["col2"]])


@materialize(lazy=True)
def simple_lazy_table_with_indexes():
    query = _get_df_query()
    return Table(query, indexes=[["col2"], ["col2", "col1"]])


@materialize(nout=2, lazy=False)
def simple_table_compressed_one_method():
    query = _get_df_query()
    return Table(
        pd.DataFrame(dict(COMPRESSION=["V"], ROWCOMPMODE=[" "])),
        "compress_1_properties",
    ), Table(query, name="compress_one", materialization_details="value_compression")


@materialize(version="1.0")
def compression_properties_adaptive_value_compression_db2():
    return Table(
        pd.DataFrame(dict(COMPRESSION=["B"], ROWCOMPMODE=["A"])),
        "compress_2_properties",
    )


@materialize(nout=2, lazy=False)
def simple_table_compressed_two_methods():
    query = _get_df_query()
    return compression_properties_adaptive_value_compression_db2(), Table(
        query, name="compress_two", materialization_details="adaptive_value_compression"
    )


@materialize(nout=2, lazy=False)
def simple_table_default_compressed():
    # The stage in test_compression has
    # materialization_details="adaptive_value_compression".
    # This justifies the use of compression_properties_adaptive_value_compression_db2().
    query = _get_df_query()
    return compression_properties_adaptive_value_compression_db2(), Table(query, name="compress_two")


@materialize(version="1.0")
def pd_dataframe(data: dict[str, list]):
    return pd.DataFrame(data)


@materialize(version="1.0")
def as_blob(x):
    return Blob(x)


@materialize(lazy=True)
def as_raw_sql(sql: str):
    return RawSql(sql)


class _SomeClass:
    def __eq__(self, other):
        return self.__dict__ == other.__dict__


@materialize(version="1.0")
def object_blob(x: dict):
    instance = _SomeClass()
    instance.__dict__.update(x)

    return Blob(instance)


@materialize()
def exception(x, r: bool):
    if r:
        raise Exception("THIS EXCEPTION IS EXPECTED")
    return x


def get_task_logger():
    return TaskContext.get().task.logger


@materialize(input_type=sa.Table)
def check_pk_length(x: sa.Table):
    pks = inspect(x).primary_key
    if isinstance(pks, list):
        pk_name = pks[0].name
        assert len(pk_name) <= MAX_LENGTH_PK



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod

from pydiverse.common.util import Disposable
from pydiverse.pipedag import ExternalTableReference, Task
from pydiverse.pipedag.core import Result, Subflow


class OrchestrationEngine(Disposable, ABC):
    """Flow orchestration engine base class"""

    @abstractmethod
    def run(
        self,
        flow: Subflow,
        ignore_position_hashes: bool = False,
        inputs: dict[Task, ExternalTableReference] | None = None,
        **kwargs,
    ) -> Result:
        """Execute a flow

        :param flow: the pipedag flow to execute
        :param ignore_position_hashes:
            If ``True``, the position hashes of tasks are not checked
            when retrieving the inputs of a task from the cache.
            This simplifies execution of subgraphs if you don't care whether inputs to
            that subgraph are cache invalid. This allows multiple modifications in the
            Graph before the next run updating the cache.
            Attention: This may break automatic cache invalidation.
            And for this to work, any task producing an input
            for the chosen subgraph may never be used more
            than once per stage.
        :param kwargs: Optional keyword arguments. How they get used is
            engine specific.
        :return: A result instance wrapping the flow execution result.
        """



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import json
import textwrap
import time
import warnings
from collections.abc import Iterable
from contextlib import contextmanager
from typing import Any, Literal

import sqlalchemy as sa
import sqlalchemy.exc

from pydiverse.common.util.hashing import stable_hash
from pydiverse.pipedag import Stage, Table
from pydiverse.pipedag.backend.table.sql.ddl import (
    AddIndex,
    AddPrimaryKey,
    ChangeColumnNullable,
    CopyTable,
    CreateAlias,
    CreateDatabase,
    CreateSchema,
    CreateTableAsSelect,
    DropAlias,
    DropSchema,
    DropSchemaContent,
    DropTable,
    RenameSchema,
    RenameTable,
    split_ddl_statement,
)
from pydiverse.pipedag.container import RawSql, Schema
from pydiverse.pipedag.context import RunContext
from pydiverse.pipedag.context.context import (
    CacheValidationMode,
    ConfigContext,
    StageCommitTechnique,
)
from pydiverse.pipedag.context.run_context import DeferredTableStoreOp
from pydiverse.pipedag.errors import CacheError
from pydiverse.pipedag.materialize.details import resolve_materialization_details_label
from pydiverse.pipedag.materialize.materializing_task import MaterializingTask
from pydiverse.pipedag.materialize.metadata import (
    LazyTableMetadata,
    RawSqlMetadata,
    TaskMetadata,
)
from pydiverse.pipedag.materialize.store import BaseTableStore

try:
    from sqlalchemy import Connection, Engine
except ImportError:
    # For compatibility with sqlalchemy < 2.0
    from sqlalchemy.engine import Engine
    from sqlalchemy.engine.base import Connection

DISABLE_DIALECT_REGISTRATION = "__DISABLE_DIALECT_REGISTRATION"


class SQLTableStore(BaseTableStore):
    """Table store that materializes tables to a SQL database

    Uses schema swapping for transactions:
    Creates a schema for each stage and a temporary schema for each
    transaction. If all tasks inside a stage succeed, swaps the schemas by
    renaming them.

    The correct dialect specific subclass of ``SQLTableStore`` gets initialized when
    based on the dialect found in the provided engine url during initialization.

    .. rubric:: Supported Tables

    The `SQLTableStore` can materialize (that is, store task output)
    and dematerialize (that is, retrieve task input) the following Table types:

    .. list-table::
       :widths: 30, 30, 30
       :header-rows: 1
       :stub-columns: 1

       * - Framework
         - Materialization
         - Dematerialization

       * - SQLAlchemy
         - | :py:class:`sa.sql.expression.Selectable
                        <sqlalchemy.sql.expression.Selectable>`
           | :py:class:`sa.sql.expression.TextClause
                        <sqlalchemy.sql.expression.TextClause>`
         - :py:class:`sa.Table <sqlalchemy.schema.Table>`

       * - Pandas
         - :py:class:`pd.DataFrame <pandas.DataFrame>`
         - :py:class:`pd.DataFrame <pandas.DataFrame>`

       * - Polars
         - | :external+pl:doc:`pl.DataFrame <reference/dataframe/index>`
           | :external+pl:doc:`pl.LazyFrame <reference/lazyframe/index>`
         - | :external+pl:doc:`pl.DataFrame <reference/dataframe/index>`
           | :external+pl:doc:`pl.LazyFrame <reference/lazyframe/index>`

       * - tidypolars
         - :py:class:`tp.Tibble <tidypolars.tibble.Tibble>`
         - :py:class:`tp.Tibble <tidypolars.tibble.Tibble>`

       * - Ibis
         - :py:class:`ibis.api.Table <ibis.expr.types.relations.Table>`
         - :py:class:`ibis.api.Table <ibis.expr.types.relations.Table>`

       * - pydiverse.transform
         - ``pdt.Table``
         - | ``pdt.eager.PandasTableImpl``
           | ``pdt.lazy.SQLTableImpl``

       * - pydiverse.pipedag table reference
         - :py:class:`~.ExternalTableReference` (no materialization)
         - Can be read with all dematerialization methods above

    :param url:
        The :external+sa:ref:`SQLAlchemy engine url <database_urls>`
        used to connect to the database.

        This URL may contain placeholders like ``{name}`` or ``{instance_id}``
        (additional ones can be defined in the ``url_attrs_file``) or
        environment variables like ``{$USER}`` which get substituted with their
        respective values.

        Attention: passwords including special characters like ``@`` or ``:`` need
        to be URL encoded.

    :param url_attrs_file:
        Filename of a yaml file which is read shortly before rendering the final
        engine URL and which is used to replace custom placeholders in ``url``.

        Just like ``url``, this value may also contain placeholders and environment
        variables which get substituted.

    :param create_database_if_not_exists:
        If the engine url references a database name that doesn't yet exists,
        then setting this value to ``True`` tells pipedag to create the database
        before trying to open a connection to it.

    :param schema_prefix:
        A prefix that gets placed in front of all schema names created by pipedag.

    :param schema_suffix:
        A suffix that gets placed behind of all schema names created by pipedag.

    :param avoid_drop_create_schema:
        If ``True``, no ``CREATE SCHEMA`` or ``DROP SCHEMA`` statements get issued.
        This is mostly relevant for databases that support automatic schema
        creation like IBM DB2.

    :param print_materialize:
        If ``True``, all tables that get materialized get logged.

    :param print_sql:
        If ``True``, all executed SQL statements get logged.

    :param no_db_locking:
        Speed up database by telling it we will not rely on it's locking mechanisms.
        Currently not implemented.

    :param strict_materialization_details:
        If ``True``: raise an exception if
            - the argument ``materialization_details`` is given even though the
              table store does not support it.
            - a table references a ``materialization_details`` tag that is not defined
              in the config.
        If ``False``: Log an error instead of raising an exception

    :param materialization_details:
        A dictionary of tags that each define properties used for materialization by
        the table store. If the tag ``__any__`` is present, the other labels will
        inherit properties from it if they do not override them.

        An example config for DB2 that goes into the ``args`` section of the
        ``table_store`` config in ``pipedag.yaml``:

        .. code-block:: yaml

            materialization_details:
              __any__:
                compression: ["COMPRESS YES ADAPTIVE", "VALUE COMPRESSION"]
                table_space_data: "USERSPACE1"
              no_compression:
                # user-defined tag. Inherits table_space_data from __any__
                # but overwrites compression.
                compression: ""

        Then, by default, all tables will be created with the ``__any__`` config,
        unless the label ``no_compression`` is specified at stage or table level.

        See the documentation of the SQLTableStore Dialects for supported options for
        each table store.

    :param default_materialization_details:
        The materialization_details that will be used if materialization_details
        is not specified on table level. If not set, the ``__any__`` tag (if specified)
        will be used.
    :param max_concurrent_copy_operations:
        In case of a partially cache-valid stage, we need to copy tables from the
        cache schema to the new transaction schema. This parameter specifies the
        maximum number of workers we use for concurrently copying tables.
    :param sqlalchemy_pool_size:
        The number of connections to keep open inside the connection pool.
        It is recommended to choose a larger number than
        ``max_concurrent_copy_operations`` to avoid running into pool_timeout.
    :param sqlalchemy_pool_timeout:
        The number of seconds to wait before giving up on getting a connection from
        the pool. This may be relevant in case the connection pool is saturated
        with concurrent operations each working with one or more database connections.
    """

    METADATA_SCHEMA = "pipedag_metadata"
    LOCK_SCHEMA = "pipedag_locks"

    __registered_dialects: dict[str, type["SQLTableStore"]] = {}
    _dialect_name: str

    def __new__(cls, engine_url: str, *args, **kwargs):
        if cls != SQLTableStore:
            return super().__new__(cls)

        # If calling SQLTableStore(engine_url), then we want to dynamically instantiate
        # the correct dialect specific subclass based on the dialect found in the url.
        dialect = sa.engine.make_url(engine_url).get_dialect().name
        dialect_specific_cls = SQLTableStore.__registered_dialects.get(dialect, cls)
        return super(SQLTableStore, dialect_specific_cls).__new__(dialect_specific_cls)

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        config = config.copy()
        engine_url = config.pop("url")
        return cls(engine_url, **config)

    def __init__(
        self,
        engine_url: str,
        *,
        create_database_if_not_exists: bool = False,
        schema_prefix: str = "",
        schema_suffix: str = "",
        avoid_drop_create_schema: bool = False,
        print_materialize: bool = False,
        print_sql: bool = False,
        no_db_locking: bool = True,
        strict_materialization_details: bool = True,
        materialization_details: dict[str, dict[str, str | list[str]]] | None = None,
        default_materialization_details: str | None = None,
        max_concurrent_copy_operations: int = 5,
        sqlalchemy_pool_size: int = 12,
        sqlalchemy_pool_timeout: int = 300,
    ):
        super().__init__()

        self.create_database_if_not_exists = create_database_if_not_exists
        self.schema_prefix = schema_prefix
        self.schema_suffix = schema_suffix
        self.avoid_drop_create_schema = avoid_drop_create_schema
        self.print_materialize = print_materialize
        self.print_sql = print_sql
        self.no_db_locking = no_db_locking
        self.strict_materialization_details = strict_materialization_details
        self.max_concurrent_copy_operations = max_concurrent_copy_operations
        self.sqlalchemy_pool_size = sqlalchemy_pool_size
        self.squalchemy_pool_timeout = sqlalchemy_pool_timeout

        self.metadata_schema = self.get_schema(self.METADATA_SCHEMA)
        self.engine_url = sa.engine.make_url(engine_url)

        self._init_database_before_engine()
        self.engine = self._create_engine()

        # Dict where hooks are allowed to store values that should get cached
        self.hook_cache = {}

        self._init_database()

        # Set up metadata tables and schema
        self.sql_metadata = sa.MetaData(schema=self.metadata_schema.get())

        # Store version number for metadata table schema evolution.
        # We disable caching in case of version mismatch.
        self.disable_caching = False
        self.metadata_version = "0.3.2"  # Increase version if metadata table changes
        self.version_table = sa.Table(
            "metadata_version",
            self.sql_metadata,
            sa.Column("version", sa.String(32)),
        )

        # Stage Table is unique for stage
        self.stage_table = sa.Table(
            "stages",
            self.sql_metadata,
            self._metadata_pk("id", "stages"),
            sa.Column("stage", sa.String(64)),
            sa.Column("cur_transaction_name", sa.String(256)),
        )

        # Task Table is unique for stage * in_transaction_schema
        self.tasks_table = sa.Table(
            "tasks",
            self.sql_metadata,
            self._metadata_pk("id", "tasks"),
            sa.Column("name", sa.String(128)),
            sa.Column("stage", sa.String(64)),
            sa.Column("version", sa.String(64)),
            sa.Column("timestamp", sa.DateTime()),
            sa.Column("run_id", sa.String(20)),
            sa.Column("position_hash", sa.String(20)),
            sa.Column("input_hash", sa.String(20)),
            sa.Column("cache_fn_hash", sa.String(20)),
            sa.Column("output_json", sa.Text()),
            sa.Column("in_transaction_schema", sa.Boolean()),
        )

        # Lazy Cache Table is unique for stage * in_transaction_schema
        self.lazy_cache_table = sa.Table(
            "lazy_tables",
            self.sql_metadata,
            self._metadata_pk("id", "lazy_tables"),
            sa.Column("name", sa.String(128)),
            sa.Column("stage", sa.String(64)),
            sa.Column("query_hash", sa.String(20)),
            sa.Column("task_hash", sa.String(20)),
            sa.Column("in_transaction_schema", sa.Boolean()),
        )

        # Sql Cache Table is unique for stage * in_transaction_schema
        self.raw_sql_cache_table = sa.Table(
            "raw_sql",
            self.sql_metadata,
            self._metadata_pk("id", "raw_sql"),
            sa.Column("prev_objects", sa.Text()),
            sa.Column("new_objects", sa.Text()),
            sa.Column("stage", sa.String(64)),
            sa.Column("query_hash", sa.String(20)),
            sa.Column("task_hash", sa.String(20)),
            sa.Column("in_transaction_schema", sa.Boolean()),
        )

        self.default_materialization_details = default_materialization_details

        self._set_materialization_details(materialization_details)

        self.logger.info(
            "Initialized SQL Table Store",
            engine_url=self.engine_url.render_as_string(hide_password=True),
            schema_prefix=self.schema_prefix,
            schema_suffix=self.schema_suffix,
        )

    def __init_subclass__(cls, **kwargs):
        # Whenever a new subclass if SQLTableStore is defined, it must contain the
        # `_dialect_name` attribute. This allows us to dynamically instantiate it
        # when calling SQLTableStore(engine_url, ...) based on the dialect name found
        # in the engine url (see __new__).
        dialect_name = getattr(cls, "_dialect_name", None)
        if dialect_name is None:
            raise ValueError(
                "All subclasses of SQLTableStore must have a `_dialect_name` attribute."
                f" But {cls.__name__}._dialect_name is None."
            )

        if dialect_name != DISABLE_DIALECT_REGISTRATION:
            if dialect_name in SQLTableStore.__registered_dialects:
                warnings.warn(f"Already registered a SQLTableStore for dialect {dialect_name}")
            SQLTableStore.__registered_dialects[dialect_name] = cls

    def _metadata_pk(self, name: str, table_name: str):
        return sa.Column(name, sa.BigInteger(), primary_key=True)

    def _init_database_before_engine(self):
        "File databases might have to create a directory here."
        pass

    def _init_database(self):
        if not self.create_database_if_not_exists:
            return

        raise NotImplementedError("create_database_if_not_exists is not implemented for this engine")

    def _init_database_with_database(
        self,
        database: str,
        execution_options: dict = None,
        create_database: str | None = None,
        disable_exists_check=False,
    ):
        if not self.create_database_if_not_exists:
            return

        if execution_options is None:
            execution_options = {}

        if not disable_exists_check:
            try:
                # Check if database exists
                with self.engine.connect() as conn:
                    conn.exec_driver_sql("SELECT 1")
                return
            except sa.exc.DBAPIError:
                # Database doesn't exist
                pass

        # Create new database using temporary engine object
        tmp_db_url = self.engine_url.set(database=database)
        tmp_engine = sa.create_engine(tmp_db_url, execution_options=execution_options)

        try:
            if create_database is None:
                create_database = self.engine_url.database
            with tmp_engine.connect() as conn:
                conn.execute(CreateDatabase(create_database))
        except sa.exc.DBAPIError:
            # This happens if multiple instances try to create the database
            # at the same time.
            with self.engine.connect() as conn:
                # Verify database actually exists
                conn.exec_driver_sql("SELECT 1")

    def _default_isolation_level(self) -> str | None:
        # we never want database transactional behavior (concurrent read+write)
        return "READ UNCOMMITTED"

    def _create_engine(self, **kwargs):
        if self._default_isolation_level() is not None and "isolation_level" not in kwargs:
            kwargs["isolation_level"] = self._default_isolation_level()

        # future=True enables SQLAlchemy 2.0 behaviour with version 1.4
        # However, it does not support pd.read_sql_table
        return sa.create_engine(
            self.engine_url,
            # future=True,
            pool_size=self.sqlalchemy_pool_size,
            pool_timeout=self.squalchemy_pool_timeout,
            **kwargs,
        )

    @contextmanager
    def engine_connect(self) -> Connection:
        with self.engine.connect() as conn:
            yield conn
            if sa.__version__ >= "2.0.0":
                conn.commit()

    def get_schema(self, name: str) -> Schema:
        return Schema(name, self.schema_prefix, self.schema_suffix)

    def _execute(self, query, conn: sa.engine.Connection, truncate_printed_select=False):
        if self.print_sql:
            if isinstance(query, str):
                query_str = query
            else:
                query_str = str(query.compile(self.engine, compile_kwargs={"literal_binds": True}))
            pretty_query_str = self.format_sql_string(query_str, truncate_printed_select)
            self.logger.info("Executing sql", query=pretty_query_str)

        if isinstance(query, str):
            ret = conn.execute(sa.text(query))
        else:
            ret = conn.execute(query)
        return ret

    def execute(
        self,
        query: str
        | sqlalchemy.sql.expression.Selectable
        | sqlalchemy.sql.expression.TextClause
        | sqlalchemy.sql.ddl.DDLElement
        | list[
            str
            | sqlalchemy.sql.expression.Selectable
            | sqlalchemy.sql.expression.TextClause
            | sqlalchemy.sql.ddl.DDLElement
        ],
        *,
        conn: sa.engine.Connection = None,
        truncate_printed_select=False,
    ):
        if conn is None:
            with self.engine_connect() as conn:
                return self.execute(
                    query,
                    conn=conn,
                    truncate_printed_select=truncate_printed_select,
                )

        if isinstance(query, sa.schema.DDLElement) or isinstance(query, list):
            if not isinstance(query, list):
                query = [query]

            @contextmanager
            def transactional():
                if conn.in_transaction():
                    yield
                else:
                    with conn.begin():
                        yield

            with transactional():
                # execute multiple statements in one transaction
                for cur_query in query:
                    if isinstance(cur_query, sa.schema.DDLElement):
                        # Some custom DDL statements contain multiple statements.
                        # They are all separated using a special separator.
                        query_str = str(cur_query.compile(self.engine, compile_kwargs={"literal_binds": True}))

                        for part in split_ddl_statement(query_str):
                            self._execute(part, conn, truncate_printed_select)
                    else:
                        self._execute(cur_query, conn, truncate_printed_select)
            return

        return self._execute(query, conn, truncate_printed_select)

    def reflect_table(self, table_name: str, schema: str | Schema) -> sa.Table:
        if isinstance(schema, Schema):
            schema = schema.get()
        tbl = None
        for retry_iteration in range(4):
            # retry operation since it might have been terminated as a deadlock victim
            try:
                tbl = sa.Table(
                    table_name,
                    sa.MetaData(),
                    schema=schema,
                    autoload_with=self.engine,
                )
                break
            except (sa.exc.SQLAlchemyError, sa.exc.DBAPIError):
                if retry_iteration == 3:
                    raise
                time.sleep(retry_iteration * retry_iteration * 1.2)
        return tbl

    def rename_table(self, table: Table, to_name: str, schema: Schema):
        """Method can be overwritten by derived table stores like ParquetTableStore."""
        self.execute(RenameTable(table.name, to_name, schema))

    def write_subquery(
        self,
        query: sa.Text | sqlalchemy.sql.expression.Selectable | sqlalchemy.sql.expression.TextClause,
        to_name: str,
        neighbor_table: Table,
        *,
        unlogged: bool = False,
        suffix: str | None = None,
    ):
        """Write a query to a table in same schema as neighbor_table.

        This is mainly for overriding by derived table stores like ParquetTableStore.
        """
        schema = self.get_schema(neighbor_table.stage.transaction_name)
        self.execute(
            CreateTableAsSelect(
                to_name,
                schema,
                query,
                unlogged=unlogged,
                suffix=suffix or "",
            )
        )
        return schema

    def drop_subquery_table(
        self,
        drop_name: str,
        schema: Schema | str,
        neighbor_table: Table,
        if_exists: bool = False,
        cascade: bool = False,
    ):
        """Drop a table in same schema as neighbor_table.

        This is mainly for overriding by derived table stores like ParquetTableStore.
        """
        _ = neighbor_table
        self.execute(DropTable(drop_name, schema, if_exists=if_exists, cascade=cascade))

    def check_materialization_details_supported(self, label: str | None) -> None:
        if label is None:
            return
        error_msg = f"Materialization details are not supported for store {type(self).__name__}."
        if self.strict_materialization_details:
            raise ValueError(f"{error_msg} To silence this exception set strict_materialization_details=False")
        else:
            self.logger.error(f"{error_msg}")

    def _set_materialization_details(self, materialization_details: dict[str, dict[str | list[str]]] | None) -> None:
        if materialization_details is not None or self.default_materialization_details is not None:
            error_msg = f"{type(self).__name__} does not support materialization details."
            if self.strict_materialization_details:
                raise TypeError(f"{error_msg} To suppress this exception, use strict_materialization_details=False")
            self.logger.error(error_msg)

    def get_unlogged(self, materialization_details_label: str | None) -> bool:
        _ = materialization_details_label
        return False  # only available for some dialects

    def get_create_table_suffix(self, materialization_details_label: str | None) -> str:
        """
        This method can be used to append materialization tables to CREATE TABLE
        statements by dialects that support it.

        :param materialization_details_label:
            A label that can be controlled on table or stage level to reference
            options laid out in detail in configuration
        :return:
            Suffix that is appended to CREATE TABLE statement as string
        """
        return ""

    def get_limit_query(
        self,
        query: sa.sql.expression.Selectable | sa.sql.expression.TextClause,
        rows: int,
    ) -> sa.sql.expression.Select:
        if isinstance(query, sa.sql.expression.TextClause):
            return sa.select(sa.text("*")).limit(rows).select_from(sa.text("(" + str(query) + ") AS A"))
        else:
            return sa.select(sa.text("*")).limit(rows).select_from(query.alias("A"))

    def lock_table(self, table: Table | str, schema: Schema | str, conn: Any = None) -> list:
        """
        For some dialects, it might be beneficial to lock a table before writing to it.

        :returns
            A list of SQLAlchemy statements that should be executed if conn=None
        """
        _ = table, schema, conn
        return []

    def lock_source_table(self, table: Table | str, schema: Schema | str, conn: Any = None) -> list:
        """
        For some dialects, it might be beneficial to lock source tables before reading.

        :returns
            A list of SQLAlchemy statements that should be executed if conn=None
        """
        _ = table, schema, conn
        return []

    def lock_source_tables(self, tables: list[dict[str, str]], conn: Any = None) -> list:
        """
        For some dialects, it might be beneficial to lock source tables before reading.

        :returns
            A list of SQLAlchemy statements that should be executed if conn=None
        """
        stmts = []
        for table in tables:
            if table["shared_lock_allowed"]:
                stmts += self.lock_source_table(table["name"], table["schema"], conn)
        return stmts

    def get_forced_nullability_columns(self, table: Table, table_cols: Iterable[str]) -> tuple[list[str], list[str]]:
        """
        Get the columns that should be forced to be nullable or non-nullable.

        This is typically applied on the materialized empty table, but may also
        be applied after filling the table by some dialects. Some dialects may
        choose to add primary key columns to non_nullable_cols if they need it
        to set the primary key. However, some dialects do this separately because
        they also need to change the type of primary key columns. Some dialects
        only return non-nullable columns because all columns will be nullable
        after initial materialization.

        :param table:
            Table to be materialized. It includes attributes like nullable,
            non_nullable, and primary_key which may influence the return value.
        :param table_cols:
            The list of all columns in the table. If nullable and non_nullable is
            specified, all columns must be mentioned. If only one of them is set,
            the other one is calculated as the remaining set.
        :return:
            A tuple of two lists. The first list contains the columns that should
            be forced to be nullable, the second list contains the columns that
            should be forced to be non-nullable. Most dialects will not need to
            force nullable columns because very column is nullable by default.
        """
        _, non_nullable_cols = self._process_table_nullable_parameters(table, table_cols)

        # in most dialects columns are nullable by default
        return [], non_nullable_cols

    @staticmethod
    def _process_table_nullable_parameters(table: Table, table_cols: Iterable[str]):
        name = f'"{table.name}"'
        nullable_set = set(table.nullable or [])
        non_nullable_set = set(table.non_nullable or [])
        table_cols_set = set(table_cols)

        for prefix, col_set, cols in [
            ("", nullable_set, table.nullable),
            ("non_", non_nullable_set, table.non_nullable),
        ]:
            if invalid_cols := col_set - table_cols_set:
                raise ValueError(
                    f"The columns {invalid_cols} in Table({name},"
                    f" {prefix}nullable={cols}) aren't contained in the table"
                    f" columns: {table_cols}"
                )

        if table.nullable is not None and table.non_nullable is not None:
            # If both are specified, they aren't allowed to intersect and
            # their union must be the set of all columns.
            if intersection := nullable_set & non_nullable_set:
                raise ValueError(
                    f"The columns {intersection} are both set to be nullable and non"
                    f" nullable in Table({name}, nullable={table.nullable},"
                    f" non_nullable={table.non_nullable})"
                )
            if missing_cols := table_cols_set - nullable_set - non_nullable_set:
                raise ValueError(
                    f"When setting Table({name}, nullable={table.nullable},"
                    f" non_nullable={table.non_nullable}), all columns of the"
                    f" table must be mentioned. Missing columns: {missing_cols}"
                )
            nullable = table.nullable
            non_nullable = table.non_nullable
        elif table.nullable is not None:
            non_nullable = [c for c in table_cols if c not in nullable_set]
            nullable = table.nullable
        elif table.non_nullable is not None:
            nullable = [c for c in table_cols if c not in non_nullable_set]
            non_nullable = table.non_nullable
        else:
            nullable = []
            non_nullable = []

        return nullable, non_nullable

    def dialect_requests_empty_creation(self, table: Table, is_sql: bool) -> bool:
        _ = is_sql
        return (
            table.nullable is not None
            or table.non_nullable is not None
            or self.get_unlogged(resolve_materialization_details_label(table))
        )

    def add_indexes_and_set_nullable(
        self,
        table: Table,
        schema: Schema,
        *,
        on_empty_table: bool | None = None,
        table_cols: Iterable[str] | None = None,
    ):
        if on_empty_table is None or on_empty_table:
            # By default, we set non-nullable on empty table
            if table_cols is None:
                # reflect table:
                inspector = sa.inspect(self.engine)
                columns = inspector.get_columns(table.name, schema=schema.get())
                table_cols = [col["name"] for col in columns]
            nullable_cols, non_nullable_cols = self.get_forced_nullability_columns(table, table_cols)
            if len(nullable_cols) > 0:
                # some dialects represent literals as non-nullable types
                self.execute(ChangeColumnNullable(table.name, schema, nullable_cols, nullable=True))
            if len(non_nullable_cols) > 0:
                self.execute(ChangeColumnNullable(table.name, schema, non_nullable_cols, nullable=False))
        if on_empty_table is None or not on_empty_table:
            # By default, we set indexes after loading full table. This can be
            # overridden by dialect
            self.add_table_primary_key(table, schema)
            self.add_table_indexes(table, schema)

    def add_table_indexes(self, table: Table, schema: Schema):
        if table.indexes is not None:
            for index in table.indexes:
                self.add_index(table.name, schema, index)

    def add_table_primary_key(self, table: Table, schema: Schema):
        if table.primary_key is not None:
            key = table.primary_key
            if isinstance(key, str):
                key = [key]
            self.add_primary_key(table.name, schema, key)

    def add_primary_key(
        self,
        table_name: str,
        schema: Schema,
        key_columns: list[str],
        *,
        name: str | None = None,
    ):
        self.execute(AddPrimaryKey(table_name, schema, key_columns, name))

    def add_index(
        self,
        table_name: str,
        schema: Schema,
        index_columns: list[str],
        name: str | None = None,
    ):
        self.execute(AddIndex(table_name, schema, index_columns, name))

    def copy_indexes(self, src_table: str, src_schema: Schema, dest_table: str, dest_schema: Schema):
        inspector = sa.inspect(self.engine)
        pk_constraint = inspector.get_pk_constraint(src_table, schema=src_schema.get())
        if len(pk_constraint["constrained_columns"]) > 0:
            self.add_primary_key(
                dest_table,
                dest_schema,
                pk_constraint["constrained_columns"],
                name=pk_constraint["name"],
            )
        indexes = inspector.get_indexes(src_table, schema=src_schema.get())
        for index in indexes:
            if len(index["include_columns"]) > 0:
                self.logger.warning(
                    "Perfect index recreation in caching is not net implemented",
                    table=dest_table,
                    schema=dest_schema.get(),
                    include_columns=index["include_columns"],
                )
            if any(not isinstance(val, list) or len(val) > 0 for val in index["dialect_options"].values()):
                self.logger.warning(
                    "Perfect index recreation in caching is not net implemented",
                    table=dest_table,
                    schema=dest_schema.get(),
                    dialect_options=index["dialect_options"],
                )
            self.add_index(dest_table, dest_schema, index["column_names"], index["name"])

    def reflect_sql_types(self, col_names: Iterable[str], table_name: str, schema: Schema):
        inspector = sa.inspect(self.engine)
        columns = inspector.get_columns(table_name, schema=schema.get())
        types = {d["name"]: d["type"] for d in columns}
        sql_types = [types[col] for col in col_names]
        return sql_types

    @staticmethod
    def format_sql_string(query_str: str, truncate_select=False) -> str:
        if truncate_select and "SELECT" in query_str:
            query_str = query_str.split("SELECT")[0] + "SELECT ...\n"
        return textwrap.dedent(query_str).strip()

    def execute_raw_sql(self, raw_sql: RawSql):
        """Executed raw SQL statements in the associated transaction stage"""
        for statement in raw_sql.sql.split(raw_sql.separator):
            if not statement.strip():
                # Skip empty queries
                continue

            self.execute(statement)

    def setup(self):
        super().setup()
        if not self.avoid_drop_create_schema:
            self.execute(CreateSchema(self.metadata_schema, if_not_exists=True))
        with self.engine_connect() as conn:
            try:
                version = conn.execute(sa.select(self.version_table.c.version)).scalar_one_or_none()
            except sa.exc.ProgrammingError:
                # table metadata_version does not yet exist
                version = None
        if version is None:
            with self.engine_connect() as conn, conn.begin():
                self.sql_metadata.create_all(conn)
                version = conn.execute(sa.select(self.version_table.c.version)).scalar_one_or_none()
                assert version is None  # detect race condition
                conn.execute(
                    self.version_table.insert().values(
                        version=self.metadata_version,
                    )
                )
        elif version != self.metadata_version:
            # disable caching due to incompatible metadata table schemas
            # (in future versions, consider automatic metadata schema upgrade)
            self.disable_caching = True
            self.logger.warning(
                "Disabled caching due to metadata version mismatch",
                version=version,
                expected_version=self.metadata_version,
            )

    def dispose(self):
        self.engine.dispose()
        self.hook_cache = None
        super().dispose()

    def init_stage(self, stage: Stage):
        stage_commit_technique = ConfigContext.get().stage_commit_technique
        if stage_commit_technique == StageCommitTechnique.SCHEMA_SWAP:
            return self._init_stage_schema_swap(stage)
        if stage_commit_technique == StageCommitTechnique.READ_VIEWS:
            return self._init_stage_read_views(stage)
        raise ValueError(f"Invalid stage commit technique: {stage_commit_technique}")

    def optional_pause_for_db_transactionality(
        self,
        prev_action: Literal[
            "table_drop",
            "table_create",
            "schema_drop",
            "schema_create",
            "schema_rename",
        ],
    ):
        _ = type
        # Some backends have transactionality problems with very quick
        # DROP/CREATE or RENAME activities for both schemas and tables
        # which happen in testing.
        pass

    def _init_stage_schema_swap(self, stage: Stage):
        schema = self.get_schema(stage.name)
        transaction_schema = self.get_schema(stage.transaction_name)

        cs_base = CreateSchema(schema, if_not_exists=True)
        ds_trans = DropSchema(transaction_schema, if_exists=True, cascade=True, engine=self.engine)
        cs_trans = CreateSchema(transaction_schema, if_not_exists=False)

        with self.engine_connect() as conn:
            if self.avoid_drop_create_schema:
                self.execute(DropSchemaContent(transaction_schema, self.engine), conn=conn)
                self.optional_pause_for_db_transactionality("table_drop")
            else:
                self.execute(cs_base, conn=conn)
                self.execute(ds_trans, conn=conn)
                self.optional_pause_for_db_transactionality("schema_drop")
                self.execute(cs_trans, conn=conn)
                self.optional_pause_for_db_transactionality("schema_create")

            if not self.disable_caching:
                for table in [
                    self.tasks_table,
                    self.lazy_cache_table,
                    self.raw_sql_cache_table,
                ]:
                    conn.execute(table.delete().where(table.c.stage == stage.name).where(table.c.in_transaction_schema))

    def _init_stage_read_views(self, stage: Stage):
        cur_transaction_name = self._get_read_view_transaction_name(stage.name)

        # swap transaction name
        suffix = "__even" if cur_transaction_name.endswith("__odd") else "__odd"
        new_transaction_name = stage.name + suffix

        # TODO: WE SHOULD NOT MODIFY THE STAGE AFTER CREATION!!!
        stage.set_transaction_name(new_transaction_name)

        # Call schema swap function with updated transaction name
        self._init_stage_schema_swap(stage)

    def _get_read_view_transaction_name(self, schema_name: str):
        """Returns transaction name where committed tables are stored."""
        try:
            with self.engine_connect() as conn:
                metadata_rows = (
                    conn.execute(self.stage_table.select().where(self.stage_table.c.stage == schema_name))
                    .mappings()
                    .one()
                )
            cur_transaction_name = metadata_rows["cur_transaction_name"]
        except (sa.exc.MultipleResultsFound, sa.exc.NoResultFound):
            cur_transaction_name = ""
        return cur_transaction_name

    def commit_stage(self, stage: Stage):
        # If the stage is 100% cache valid, then we just need to update the
        # "in_transaction_schema" column of the metadata tables.
        if not RunContext.get().has_stage_changed(stage):
            self.logger.info("Stage is cache valid", stage=stage)
            with self.engine_connect() as conn:
                self._commit_stage_update_metadata(stage, conn)
            self._committed_unchanged(stage)
            return

        # Commit normally
        stage_commit_technique = ConfigContext.get().stage_commit_technique
        if stage_commit_technique == StageCommitTechnique.SCHEMA_SWAP:
            return self._commit_stage_schema_swap(stage)
        if stage_commit_technique == StageCommitTechnique.READ_VIEWS:
            return self._commit_stage_read_views(stage)
        raise ValueError(f"Invalid stage commit technique: {stage_commit_technique}")

    def _committed_unchanged(self, stage):
        # an opportunity for derived classes to take actions
        pass

    def _get_read_view_original_transaction_name(self, stage, override_transaction_name: str | None = None):
        transaction_name = stage.transaction_name if override_transaction_name is None else override_transaction_name
        # undo transaction name assignment
        suffix = "__even" if transaction_name.endswith("__odd") else "__odd"
        original_transaction_name = stage.name + suffix
        return original_transaction_name

    def _commit_stage_schema_swap(self, stage: Stage):
        schema = self.get_schema(stage.name)
        transaction_schema = self.get_schema(stage.transaction_name)
        tmp_schema = self.get_schema(stage.name + "__swap")

        # potentially this disposal must be optional since it does not allow
        # for multithreaded stage execution
        # dispose open connections which may prevent schema swapping
        self.engine.dispose()

        # We might want to also append ', conn.begin()' to the with statement,
        # but this collides with the inner conn.begin() used to execute
        # statement parts as one transaction. This is solvable via complex code,
        # but we typically don't want to rely on database transactions anyways.
        with self.engine_connect() as conn:
            # TODO: self.avoid_drop_create_schema is currently being ignored...
            self.execute(
                DropSchema(tmp_schema, if_exists=True, cascade=True, engine=self.engine),
                conn=conn,
            )
            self.optional_pause_for_db_transactionality("schema_drop")
            self.execute(RenameSchema(schema, tmp_schema, self.engine), conn=conn)
            self.optional_pause_for_db_transactionality("schema_rename")
            self.execute(RenameSchema(transaction_schema, schema, self.engine), conn=conn)
            self.optional_pause_for_db_transactionality("schema_rename")
            self.execute(RenameSchema(tmp_schema, transaction_schema, self.engine), conn=conn)
            self.optional_pause_for_db_transactionality("schema_rename")

            self._commit_stage_update_metadata(stage, conn=conn)

    def _commit_stage_read_views(self, stage: Stage):
        dest_schema = self.get_schema(stage.name)
        src_schema = self.get_schema(stage.transaction_name)

        with self.engine_connect() as conn:
            # Clear contents of destination schema
            self.execute(DropSchemaContent(dest_schema, self.engine), conn=conn)
            self.optional_pause_for_db_transactionality("table_drop")

            # Create aliases for all tables in transaction schema
            inspector = sa.inspect(self.engine)
            for table in inspector.get_table_names(schema=src_schema.get()) + inspector.get_view_names(
                schema=src_schema.get()
            ):
                self.execute(CreateAlias(table, src_schema, table, dest_schema))

            # Update metadata
            stage_metadata_exists = (
                conn.execute(sa.select(1).where(self.stage_table.c.stage == stage.name)).scalar() == 1
            )

            if stage_metadata_exists:
                conn.execute(
                    self.stage_table.update()
                    .where(self.stage_table.c.stage == stage.name)
                    .values(cur_transaction_name=stage.transaction_name)
                )
            else:
                conn.execute(
                    self.stage_table.insert().values(
                        stage=stage.name,
                        cur_transaction_name=stage.transaction_name,
                    )
                )

            self._commit_stage_update_metadata(stage, conn=conn)

    def _commit_stage_update_metadata(self, stage: Stage, conn: sa.engine.Connection):
        if not self.disable_caching:
            for table in [
                self.tasks_table,
                self.lazy_cache_table,
                self.raw_sql_cache_table,
            ]:
                conn.execute(table.delete().where(table.c.stage == stage.name).where(~table.c.in_transaction_schema))
                conn.execute(table.update().where(table.c.stage == stage.name).values(in_transaction_schema=False))

    def copy_table_to_transaction(self, table: Table):
        from_schema = self.get_schema(table.stage.name)
        from_name = table.name

        if RunContext.get().has_stage_changed(table.stage):
            self._copy_table(table, from_schema, from_name)
        else:
            self._deferred_copy_table(table, from_schema, from_name)

    def copy_lazy_table_to_transaction(self, metadata: LazyTableMetadata, table: Table):
        from_schema = self.get_schema(metadata.stage)
        from_name = metadata.name

        if RunContext.get().has_stage_changed(table.stage):
            self._copy_table(table, from_schema, from_name)
        else:
            self._deferred_copy_table(table, from_schema, from_name)

    def _copy_table(self, table: Table, from_schema: Schema, from_name: str):
        """Copies the table immediately"""
        has_table = self.has_table_or_view(from_name, schema=from_schema)
        if not has_table:
            inspector = sa.inspect(self.engine)
            available_tables = inspector.get_table_names(from_schema.get()) + inspector.get_view_names(
                from_schema.get()
            )
            msg = (
                f"Can't copy table '{from_name}' (schema: '{from_schema}') to "
                f"transaction because no such table exists.\n"
                f"Tables in schema: {available_tables}"
            )
            self.logger.error(msg)
            raise CacheError(msg)

        from_tbl = sa.Table(from_name, sa.MetaData(), schema=from_schema.get())
        dest_tbl = table.copy_without_obj()
        dest_tbl.obj = from_tbl
        dest_tbl.shared_lock_allowed = True

        # Copy table via executing a select statement with respective hook
        RunContext.get().trace_hook.cache_pre_transfer(dest_tbl)
        hook = self.get_m_table_hook(dest_tbl)
        hook.materialize(self, dest_tbl, dest_tbl.stage.transaction_name, without_config_context=True)
        RunContext.get().trace_hook.cache_post_transfer(dest_tbl)

    def _deferred_copy_table(
        self,
        table: Table,
        from_schema: Schema,
        from_name: str,
    ):
        """Tries to perform a deferred copy

        As long as the stage is determined to still be 100% cache valid, we don't
        actually copy any tables to the transaction, and instead just create an alias
        from the main schema to the transaction schema.

        If at the end of the stage, the all tables in the stage are cache valid, then
        we don't actually need to commit the stage, and instead just rename the
        tables from their old names to their new names.

        Otherwise, we copy the tables to the transaction schema in the background,
        and once we're ready to commit, we replace the aliases with the copied tables.
        """
        assert from_schema == self.get_schema(table.stage.name)

        has_table = self.has_table_or_view(from_name, from_schema)
        if not has_table:
            inspector = sa.inspect(self.engine)
            available_tables = inspector.get_table_names(from_schema.get()) + inspector.get_view_names(
                from_schema.get()
            )
            msg = (
                f"Can't deferred copy table '{from_name}' (schema: '{from_schema}') to "
                f"transaction because no such table exists.\n"
                f"Tables in schema: {available_tables}"
            )
            self.logger.error(msg)
            raise CacheError(msg)

        try:
            ctx = RunContext.get()
            stage = table.stage

            self.execute(
                CreateAlias(
                    from_name,
                    from_schema,
                    table.name,
                    self.get_schema(stage.transaction_name),
                )
            )

            table_copy = table.copy_without_obj()
            table_copy.name = f"{table.name}__copy"

            ctx.defer_table_store_op(
                stage,
                DeferredTableStoreOp(
                    "_copy_table",
                    DeferredTableStoreOp.Condition.ON_STAGE_CHANGED,
                    (
                        table_copy,
                        from_schema,
                        from_name,
                    ),
                ),
            )
            ctx.defer_table_store_op(
                stage,
                DeferredTableStoreOp(
                    "_swap_alias_with_table_copy",
                    DeferredTableStoreOp.Condition.ON_STAGE_COMMIT,
                    (),
                    {"table": table, "table_copy": table_copy},
                ),
            )
            ctx.defer_table_store_op(
                stage,
                DeferredTableStoreOp(
                    "_rename_table",
                    DeferredTableStoreOp.Condition.ON_STAGE_ABORT,
                    (from_name, table.name, from_schema),
                ),
            )

        except Exception as _e:
            msg = f"Failed to copy table {from_name} (schema: '{from_schema}') to transaction."
            self.logger.exception(msg)
            raise CacheError(msg) from _e

    def has_table_or_view(self, name, schema: Schema | str):
        if isinstance(schema, Schema):
            schema = schema.get()
        inspector = sa.inspect(self.engine)
        has_table = inspector.has_table(name, schema=schema)
        # workaround for sqlalchemy backends that fail to find views with has_table
        # in particular DuckDB https://github.com/duckdb/duckdb/issues/10322
        if not has_table:
            has_table = name in inspector.get_view_names(schema=schema)
        return has_table

    def _swap_alias_with_table_copy(self, table: Table, table_copy: Table):
        assert table_copy.stage.name == table.stage.name

        schema = self.get_schema(table.stage.transaction_name)
        try:
            self.execute(
                DropAlias(
                    table.name,
                    schema,
                )
            )
            self.execute(
                RenameTable(
                    table_copy.name,
                    table.name,
                    schema,
                )
            )
        except Exception as _e:
            msg = f"Failed putting copied table {table_copy.name} (schema: '{schema}') in place of alias."
            raise RuntimeError(msg) from _e

    def _rename_table(self, from_name: str, to_name: str, schema: Schema):
        if from_name == to_name:
            return
        self.logger.info("RENAME", fr=from_name, to=to_name)
        self.execute(RenameTable(from_name, to_name, schema))

    def copy_raw_sql_tables_to_transaction(self, metadata: RawSqlMetadata, target_stage: Stage):
        inspector = sa.inspect(self.engine)
        src_schema = self.get_schema(metadata.stage)
        dest_schema = self.get_schema(target_stage.transaction_name)

        # New tables (AND other objects)
        new_objects = set(metadata.new_objects) - set(metadata.prev_objects)
        tables_in_schema = set(inspector.get_table_names(src_schema.get()))
        objects_in_schema = self._get_all_objects_in_schema(src_schema)

        self.check_materialization_details_supported(target_stage.materialization_details)

        tables_to_copy = new_objects & tables_in_schema
        objects_to_copy = {k: v for k, v in objects_in_schema.items() if k in new_objects and k not in tables_to_copy}

        try:
            with self.engine_connect() as conn:
                for name in tables_to_copy:
                    self.execute(
                        CopyTable(
                            name,
                            src_schema,
                            name,
                            dest_schema,
                            unlogged=self.get_unlogged(target_stage.materialization_details),
                            suffix=self.get_create_table_suffix(target_stage.materialization_details),
                        ),
                        conn=conn,
                    )
                    self.copy_indexes(
                        name,
                        src_schema,
                        name,
                        dest_schema,
                    )

                for name, obj_metadata in objects_to_copy.items():
                    self._copy_object_to_transaction(name, obj_metadata, src_schema, dest_schema, conn)

        except Exception as _e:
            msg = f"Failed to copy raw sql generated object {name} (schema: '{src_schema}') to transaction."
            self.logger.exception(msg + " This error is treated as cache-lookup-failure.")
            raise CacheError(msg) from _e

    def _get_all_objects_in_schema(self, schema: Schema) -> dict[str, Any]:
        """List all objects that are present in a schema.

        This may include things like views, stored procedures, functions, etc.

        :param schema: the schema
        :return: a dictionary where the keys are the names of the objects, and the
            values are any other (dialect specific) information that might be useful.
        """

        # Because SQLAlchemy only provides limited capabilities to inspect objects
        # inside a schema, we limit ourselves to tables and views.
        inspector = sa.inspect(self.engine)
        tables = inspector.get_table_names(schema.get())
        views = inspector.get_view_names(schema.get())
        return {name: None for name in tables + views}

    def _copy_object_to_transaction(
        self,
        name: str,
        metadata: Any,
        src_schema: Schema,
        dest_schema: Schema,
        conn: Connection,
    ):
        """Copy a generic object from one schema to another.

        :param name: name of the object to be copied
        :param metadata:
            additional information as returned by `get_all_objects_in_schema`
        :param src_schema: the schema in which the object is
        :param dest_schema: the schema into which the object should be copied
        """
        dialect = self.engine.dialect.name
        raise NotImplementedError(f"Not implemented for dialect '{dialect}'.")

    def delete_table_from_transaction(self, table: Table, *, schema: Schema | None = None):
        if schema is None:
            schema = self.get_schema(table.stage.transaction_name)
        self.execute(
            DropTable(
                table.name,
                schema,
                if_exists=True,
            )
        )

    def store_task_metadata(self, metadata: TaskMetadata, stage: Stage):
        if not self.disable_caching:
            with self.engine_connect() as conn:
                conn.execute(
                    self.tasks_table.insert().values(
                        name=metadata.name,
                        stage=metadata.stage,
                        version=metadata.version,
                        timestamp=metadata.timestamp,
                        run_id=metadata.run_id,
                        position_hash=metadata.position_hash,
                        input_hash=metadata.input_hash,
                        cache_fn_hash=metadata.cache_fn_hash,
                        output_json=metadata.output_json,
                        in_transaction_schema=True,
                    )
                )

    def retrieve_task_metadata(self, task: MaterializingTask, input_hash: str, cache_fn_hash: str) -> TaskMetadata:
        if self.disable_caching:
            raise CacheError(f"Caching is disabled, so we also don't even try to retrieve task cache: {task}")

        ignore_cache_function = ConfigContext.get().cache_validation.mode == CacheValidationMode.IGNORE_FRESH_INPUT
        try:
            with self.engine_connect() as conn:
                result = (
                    conn.execute(
                        self.tasks_table.select()
                        .where(self.tasks_table.c.name == task.name)
                        .where(self.tasks_table.c.stage == task.stage.name)
                        .where(self.tasks_table.c.version == task.version)
                        .where(self.tasks_table.c.input_hash == input_hash)
                        .where(
                            self.tasks_table.c.cache_fn_hash == cache_fn_hash
                            if not ignore_cache_function
                            else sa.literal(True)
                        )
                        .where(self.tasks_table.c.in_transaction_schema.in_([False]))
                    )
                    .mappings()
                    .one_or_none()
                )
        except sa.exc.MultipleResultsFound:
            raise CacheError("Multiple results found task metadata") from None

        if result is None:
            raise CacheError(f"Couldn't retrieve task from cache: {task}")

        return TaskMetadata(
            name=result.name,
            stage=result.stage,
            version=result.version,
            timestamp=result.timestamp,
            run_id=result.run_id,
            position_hash=result.position_hash,
            input_hash=result.input_hash,
            cache_fn_hash=result.cache_fn_hash,
            output_json=result.output_json,
        )

    def retrieve_all_task_metadata(
        self, task: MaterializingTask, ignore_position_hashes: bool = False
    ) -> list[TaskMetadata]:
        match_condition = sa.and_(
            self.tasks_table.c.name == task.name,
            self.tasks_table.c.stage == task.stage.name,
        )

        if not ignore_position_hashes:
            match_condition = match_condition & (self.tasks_table.c.position_hash == task.position_hash)
        with self.engine_connect() as conn:
            results = conn.execute(self.tasks_table.select().where(match_condition)).mappings().all()
        return [
            TaskMetadata(
                name=result.name,
                stage=result.stage,
                version=result.version,
                timestamp=result.timestamp,
                run_id=result.run_id,
                position_hash=result.position_hash,
                input_hash=result.input_hash,
                cache_fn_hash=result.cache_fn_hash,
                output_json=result.output_json,
            )
            for result in results
        ]

    def store_lazy_table_metadata(self, metadata: LazyTableMetadata):
        if not self.disable_caching:
            with self.engine_connect() as conn:
                conn.execute(
                    self.lazy_cache_table.insert().values(
                        name=metadata.name,
                        stage=metadata.stage,
                        query_hash=metadata.query_hash,
                        task_hash=metadata.task_hash,
                        in_transaction_schema=True,
                    )
                )

    # noinspection DuplicatedCode
    def retrieve_lazy_table_metadata(self, query_hash: str, task_hash: str, stage: Stage) -> LazyTableMetadata:
        if self.disable_caching:
            raise CacheError("Caching is disabled, so we also don't even try to retrieve lazy table cache")

        try:
            with self.engine_connect() as conn:
                result = (
                    conn.execute(
                        self.lazy_cache_table.select()
                        .where(self.lazy_cache_table.c.stage == stage.name)
                        .where(self.lazy_cache_table.c.query_hash == query_hash)
                        .where(self.lazy_cache_table.c.task_hash == task_hash)
                        .where(self.lazy_cache_table.c.in_transaction_schema.in_([False]))
                    )
                    .mappings()
                    .one_or_none()
                )
        except sa.exc.MultipleResultsFound:
            raise CacheError("Multiple results found for lazy table cache key") from None

        if result is None:
            raise CacheError("No result found for lazy table cache key")

        return LazyTableMetadata(
            name=result.name,
            stage=result.stage,
            query_hash=result.query_hash,
            task_hash=result.task_hash,
        )

    def store_raw_sql_metadata(self, metadata: RawSqlMetadata):
        if not self.disable_caching:
            with self.engine_connect() as conn:
                conn.execute(
                    self.raw_sql_cache_table.insert().values(
                        prev_objects=json.dumps(metadata.prev_objects),
                        new_objects=json.dumps(metadata.new_objects),
                        stage=metadata.stage,
                        query_hash=metadata.query_hash,
                        task_hash=metadata.task_hash,
                        in_transaction_schema=True,
                    )
                )

    # noinspection DuplicatedCode
    def retrieve_raw_sql_metadata(self, query_hash: str, task_hash: str, stage: Stage) -> RawSqlMetadata:
        if self.disable_caching:
            raise CacheError("Caching is disabled, so we also don't even try to retrieve raw sql cache")

        try:
            with self.engine_connect() as conn:
                result = (
                    conn.execute(
                        self.raw_sql_cache_table.select()
                        .where(self.raw_sql_cache_table.c.stage == stage.name)
                        .where(self.raw_sql_cache_table.c.query_hash == query_hash)
                        .where(self.raw_sql_cache_table.c.task_hash == task_hash)
                        .where(self.raw_sql_cache_table.c.in_transaction_schema.in_([False]))
                    )
                    .mappings()
                    .one_or_none()
                )
        except sa.exc.MultipleResultsFound:
            raise CacheError("Multiple results found for raw sql cache key") from None

        if result is None:
            raise CacheError("No result found for raw sql cache key")

        return RawSqlMetadata(
            prev_objects=json.loads(result.prev_objects),
            new_objects=json.loads(result.new_objects),
            stage=result.stage,
            query_hash=result.query_hash,
            task_hash=result.task_hash,
        )

    def resolve_alias(self, table: Table, stage_name: str | None) -> tuple[str, str]:
        """
        Convert Table objects to string table name and schema.

        Take care of ExternalTableReference objects that turn into Table objects
        with external_schema not None. For normal Table objects, the stage
        schema name is needed. It will be prefixed/postfixed based on config.
        Dialect specific implementations will also try to resolve table aliases
        or synonyms since they may cause trouble with SQLAlchemy.

        :param table: Table object
        :param stage_name: Stage name component to schema (will be different
            naming pattern for currently processed transaction stage)
        :return: string table name and schema
        """
        schema = self.get_schema(stage_name).get() if table.external_schema is None else table.external_schema
        return table.name, schema

    def get_objects_in_stage(self, stage: Stage):
        schema = self.get_schema(stage.current_name)
        return list(self._get_all_objects_in_schema(schema).keys())

    def get_table_objects_in_stage(self, stage: Stage, include_views=True):
        schema = self.get_schema(stage.current_name).get()
        inspector = sa.inspect(self.engine)

        tables = inspector.get_table_names(schema)
        if not include_views:
            return tables
        views = inspector.get_view_names(schema)
        return [*tables, *views]

    def get_stage_hash(self, stage: Stage) -> str:
        """Compute hash that represents entire stage's output metadata."""
        # We only need to look in tasks_table since the output_json column is updated
        # after evaluating lazy output objects for cache validity. This is the same
        # information we use for producing input_hash of downstream tasks.
        if self.disable_caching:
            raise NotImplementedError("computing stage hash with disabled caching is currently not supported")
        with self.engine_connect() as conn:
            result = conn.execute(
                sa.select(self.tasks_table.c.output_json)
                .where(self.tasks_table.c.stage == stage.name)
                .where(self.tasks_table.c.in_transaction_schema.in_([False]))
            ).scalars()
            result = sorted(result)

        return stable_hash(*result)

    # DatabaseLockManager

    def get_engine_for_locking(self) -> Engine:
        return self._create_engine()

    def get_lock_schema(self) -> Schema:
        return self.get_schema(self.LOCK_SCHEMA)


# Load SQLTableStore Hooks
import pydiverse.pipedag.backend.table.sql.hooks  # noqa

# Load SQLTableStore dialect specific subclasses
import pydiverse.pipedag.backend.table.sql.dialects  # noqa



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from itertools import chain

import pytest

from pydiverse.pipedag import PipedagConfig

__all__ = [
    "DATABASE_INSTANCES",
    "ORCHESTRATION_INSTANCES",
    "ALL_INSTANCES",
    "with_instances",
    "skip_instances",
]


# Pytest markers associated with specific instance name
INSTANCE_MARKS = {
    # Database Instances
    "postgres": pytest.mark.postgres,
    "postgres_unlogged": pytest.mark.postgres,
    "mssql": pytest.mark.mssql,
    "mssql_pytsql": pytest.mark.mssql,
    "mssql_nobulk": pytest.mark.mssql,
    "mssql_noaodbc": pytest.mark.mssql,
    "mssql_columnstore": pytest.mark.mssql,
    "ibm_db2": pytest.mark.ibm_db2,
    "ibm_db2_avoid_schema": pytest.mark.ibm_db2,
    "ibm_db2_materialization_details": pytest.mark.ibm_db2,
    "duckdb": pytest.mark.duckdb,
    "snowflake": pytest.mark.snowflake,
    # Local Table Cache Instances
    "local_table_cache": pytest.mark.postgres,
    "local_table_cache_inout": pytest.mark.postgres,
    "local_table_cache_inout_numpy": pytest.mark.postgres,
    "local_table_store": pytest.mark.postgres,
    # Orchestration Instances
    "dask_engine": [pytest.mark.dask, pytest.mark.postgres],
    "prefect_engine": [pytest.mark.prefect, pytest.mark.postgres],
}

# Collection of instances that represent different database technologies
DATABASE_INSTANCES = (
    "postgres",
    "mssql",
    "ibm_db2",
    "duckdb",
    "parquet_backend",
    "parquet_s3_backend",
)

ORCHESTRATION_INSTANCES = (
    "dask_engine",
    "prefect_engine",
)

# Extended collection of instances
ALL_INSTANCES = (
    "postgres",
    "postgres_unlogged",
    "mssql",
    "mssql_pytsql",
    "mssql_nobulk",
    "mssql_noaodbc",
    "mssql_columnstore",
    "ibm_db2",
    "ibm_db2_avoid_schema",
    "ibm_db2_materialization_details",
    "duckdb",
    "snowflake",
    "local_table_cache",
    "parquet_backend",
    "parquet_s3_backend",
)


def with_instances(*instances, **kwargs):
    """Decorator to run a test with a specific set of instances

    :param instances: Names of the instances to use.
    :param kwargs: keyword arguments passed to PipedagConfig.default.get()
    """
    return pytest.mark.instances(*flatten(instances), **kwargs)


def skip_instances(*instances):
    """Decorator to skip running a test with a specific set of instances"""
    return pytest.mark.skip_instances(*flatten(instances))


def flatten(it):
    """Flatten an iterable"""
    if isinstance(it, (list, tuple)):
        yield from chain(*map(flatten, it))
    else:
        yield it


# FIXTURE IMPLEMENTATION


@pytest.fixture(autouse=True, scope="function", name="run_with_instance")
def fixture_run_with_instance(request):
    """Fixture that runs test with different config instances"""
    if hasattr(request, "param"):
        instance, kwargs = request.param
        config = PipedagConfig.default.get(instance=instance, **kwargs)
        with config:
            yield instance
    else:
        yield None



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import tempfile

from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import Flow, GroupNode, Stage, VisualizationStyle, materialize
from pydiverse.pipedag.core.config import create_basic_pipedag_config


@materialize
def failed():
    raise AssertionError("This task is supposed to fail")


@materialize(version=None)
def completed_but_cache_invalid():
    return 1


@materialize(version="1.0")
def cache_valid():
    return 2


@materialize(version="1.0")
def cache_valid2():
    return 3


@materialize
def skipped(out):
    return out + 1


def main():
    with tempfile.TemporaryDirectory() as temp_dir:
        cfg = (
            create_basic_pipedag_config(
                f"duckdb:///{temp_dir}/db.duckdb",
                disable_stage_locking=True,  # This is special for duckdb
                # Attention: stage and task names might be sent to the
                #   following URL. You can self-host kroki if you like:
                #   https://docs.kroki.io/kroki/setup/install/
                #   You need to install optional dependency 'pydot' for any visualization
                #   URL to appear.
                kroki_url="https://kroki.io",
                fail_fast=False,
            )
            .get("default")
            .evolve(swallow_exceptions=True)
        )
        with cfg:
            with Flow() as flow:
                with Stage("stage1"):
                    _ = completed_but_cache_invalid()
                    _ = cache_valid()
                with Stage("stage2"):
                    out = failed()
                with Stage("stage3"):
                    _ = skipped(out)
                    with GroupNode(
                        "group_none_cache_valid",
                        style=VisualizationStyle(hide_content=True),
                    ):
                        _ = completed_but_cache_invalid()
                    with GroupNode(
                        "group_any_cache_valid",
                        style=VisualizationStyle(hide_content=True),
                    ):
                        _ = completed_but_cache_invalid()
                        _ = cache_valid()
                    with GroupNode(
                        "group_all_cache_valid",
                        style=VisualizationStyle(hide_content=True),
                    ):
                        # avoid memoization (not counted as cache valid)
                        _ = cache_valid2()
                    with GroupNode("group_any_failed", style=VisualizationStyle(hide_content=True)):
                        _ = completed_but_cache_invalid()
                        out = failed()
                    with GroupNode("group_all_skipped", style=VisualizationStyle(hide_content=True)):
                        _ = skipped(out)

            # Run flow
            result = flow.run()
            assert not result.successful

            # Run flow again for cache validity
            result = flow.run()
            assert not result.successful

            # you can also visualize the flow explicitly:
            # kroki_url = result.visualize_url()
            # result.visualize()


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish
    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import time
import warnings
from typing import Literal

from pydiverse.pipedag.backend.table.sql import hooks
from pydiverse.pipedag.backend.table.sql.hooks import (
    IbisTableHook,
)
from pydiverse.pipedag.backend.table.sql.sql import SQLTableStore

try:
    import snowflake
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    snowflake = None


class SnowflakeTableStore(SQLTableStore):
    """
    SQLTableStore that supports `Snowflake`_.

    Takes the same arguments as
    :py:class:`SQLTableStore <pydiverse.pipedag.backend.table.SQLTableStore>`
    """

    _dialect_name = "snowflake"

    def _default_isolation_level(self) -> str | None:
        return None  # "READ UNCOMMITTED" does not exist in Snowflake

    def optional_pause_for_db_transactionality(
        self,
        prev_action: Literal[
            "table_drop",
            "table_create",
            "schema_drop",
            "schema_create",
            "schema_rename",
        ],
    ):
        _ = prev_action
        # The snowflake backend has transactionality problems with very quick
        # DROP/CREATE or RENAME activities for both schemas and tables
        # which happen in testing.
        time.sleep(2)

    def _init_database(self):
        create_database = self.engine_url.database.split("/")[0]
        with self.engine.connect() as conn:
            if not [
                x.name
                for x in conn.exec_driver_sql("SHOW DATABASES").mappings().all()
                if x.name.upper() == create_database.upper()
            ]:
                self._init_database_with_database(
                    "snowflake",
                    disable_exists_check=True,
                    create_database=create_database,
                )


@SnowflakeTableStore.register_table(snowflake)
class PolarsTableHook(hooks.PolarsTableHook):
    @classmethod
    def dialect_supports_connectorx(cls):
        # ConnectorX (used by Polars read_database_uri) supports Snowflake.
        return True


try:
    import ibis
except ImportError:
    ibis = None


@SnowflakeTableStore.register_table(ibis)
class IbisTableHook(IbisTableHook):
    @classmethod
    def _conn(cls, store: SnowflakeTableStore):
        return ibis.snowflake._from_url(store.engine_url)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import tempfile

import pandas as pd
import sqlalchemy as sa

from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import Flow, Stage, Table, materialize
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.core.config import create_basic_pipedag_config


@materialize(lazy=True)
def lazy_task_1():
    return Table(
        sa.select(
            sa.literal(1).label("x"),
            sa.literal(2).label("y"),
        )
    ).materialize()


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_2(input1: sa.sql.expression.Alias, input2: sa.sql.expression.Alias):
    query = sa.select(
        (input1.c.x * 5).label("x5"),
        input2.c.a,
    ).select_from(input1.outerjoin(input2, input2.c.x == input1.c.x))

    return Table(query, name="task_2_out", primary_key=["a"]).materialize()


def ref(tbl: sa.sql.expression.Alias):
    return f'"{tbl.original.schema}"."{tbl.original.name}"'


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_3(input1: sa.sql.expression.Alias):
    return Table(sa.text(f"SELECT * FROM {ref(input1)}")).materialize()


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_4(input1: sa.sql.expression.Alias):
    # imperatively materialize a subquery
    subquery = f"""
        SELECT input1.a, sum(input1.x5) as x_sum FROM {ref(input1)} as input1
        GROUP BY a
    """
    sub_ref = Table(sa.text(subquery), name="_sub").materialize()
    query = f"""
        SELECT * FROM {ref(input1)} as input1
        LEFT JOIN {ref(sub_ref)} as sub ON input1.a = sub.a
    """
    return Table(sa.text(query), name="enriched_aggregation").materialize()


@materialize(nout=2, version="1.0.0")
def eager_inputs():
    dfA = pd.DataFrame(
        {
            "a": [0, 1, 2, 4],
            "b": [9, 8, 7, 6],
        }
    )
    dfB = pd.DataFrame(
        {
            "a": [2, 1, 0, 1],
            "x": [1, 1, 2, 2],
        }
    )
    return Table(dfA, "dfA").materialize(), Table(dfB, "dfB_%%").materialize()


@materialize(version="1.0.0", input_type=pd.DataFrame)
def eager_task(tbl1: pd.DataFrame, tbl2: pd.DataFrame):
    return Table(tbl1.merge(tbl2, on="x")).materialize()


def main():
    with tempfile.TemporaryDirectory() as temp_dir:
        cfg = create_basic_pipedag_config(
            f"duckdb:///{temp_dir}/db.duckdb",
            disable_stage_locking=True,  # This is special for duckdb
            # Attention: If uncommented, stage and task names might be sent to the
            #   following URL. You can self-host kroki if you like:
            #   https://docs.kroki.io/kroki/setup/install/
            #   You need to install optional dependency 'pydot' for any visualization
            #   URL to appear.
            # kroki_url="https://kroki.io",
        ).get("default")
        with cfg:
            with Flow() as f:
                with Stage("stage_1"):
                    lazy_1 = lazy_task_1()
                    a, b = eager_inputs()

                with Stage("stage_2"):
                    lazy_2 = lazy_task_2(lazy_1, b)
                    lazy_3 = lazy_task_3(lazy_2)
                    eager = eager_task(lazy_1, b)

                with Stage("stage_3"):
                    lazy_4 = lazy_task_4(lazy_2)
                _ = lazy_3, lazy_4, eager  # unused terminal output tables

            # Run flow
            result = f.run()
            assert result.successful

            # Run in a different way for testing
            with StageLockContext():
                result = f.run()
                assert result.successful
                assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish
    main()
